#!/usr/bin/env python

import io
import unittest
import gemmi
try:
    from Bio import SeqIO
except ImportError:
    SeqIO = None

# https://rest.uniprot.org/uniprotkb/P00698.fasta
FASTA1 = """\
>sp|P00698|LYSC_CHICK Lysozyme C OS=Gallus gallus OX=9031 GN=LYZ PE=1 SV=1
MRSLLILVLCFLPLAALGKVFGRCELAAAMKRHGLDNYRGYSLGNWVCAAKFESNFNTQA
TNRNTDGSTDYGILQINSRWWCNDGRTPGSRNLCNIPCSALLSSDITASVNCAKKIVSDG
NGMNAWVAWRNRCKGTDVQAWIRGCRL
"""  # noqa: W291 - trailing whitespace

# https://en.wikipedia.org/wiki/FASTA_format
FASTA2 = """\
>MCHU - Calmodulin - Human, rabbit, bovine, rat, and chicken
MADQLTEEQIAEFKEAFSLFDKDGDGTITTKELGTVMRSLGQNPTEAELQDMINEVDADGNGTID
FPEFLTMMARKMKDTDSEEEIREAFRVFDKDGNGYISAAELRHVMTNLGEKLTDEEVDEMIREA
DIDGDGQVNYEEFVQMMTAK*

>gi|5524211|gb|AAD44166.1| cytochrome b [Elephas maximus maximus]
LCLYTHIGRNIYYGSYLYSETWNTGIMLLLITMATAFMGYVLPWGQMSFWGATVITNLFSAIPYIGTNLV
EWIWGGFSVDKATLNRFFAFHFILPFTMVALAGVHLTFLHETGSNNPLGLTSDSDKIPFHPYYTIKDFLG
LLILILLLLLLALLSPDMLGDPDNHMPADPLNTPLHIKPEWYFLFAYAILRSVPNKLGGVLALFLSIVIL
GLMPFLHTSKHRSMMLRPLSQALFWTLTMDLLTLTWIGSQPVEYPYTIIGQMASILYFSIILAFLPIAGX
IENY
"""  # noqa: W291 - trailing whitespace

# https://www.metagenomics.wiki/tools/fastq/multi-fasta-format
FASTA3 = """\
>sequenceID-001 description
AAGTAGGAATAATATCTTATCATTATAGATAAAAACCTTCTGAATTTGCTTAGTGTGTAT
ACGACTAGACATATATCAGCTCGCCGATTATTTGGATTATTCCCTG
>sequenceID-002 description
CAGTAAAGAGTGGATGTAAGAACCGTCCGATCTACCAGATGTGATAGAGGTTGCCAGTAC
AAAAATTGCATAATAATTGATTAATCCTTTAATATTGTTTAGAATATATCCGTCAGATAA
TCCTAAAAATAACGATATGATGGCGGAAATCGTC
>sequenceID-003 description
CTTCAATTACCCTGCTGACGCGAGATACCTTATGCATCGAAGGTAAAGCGATGAATTTAT
CCAAGGTTTTAATTTG
"""  # noqa: W291 - trailing whitespace

# from _entity_poly.pdbx_seq_one_letter_code from 5I55 and 1PFE
FASTA4 = """\
>
(MSE)EFVAKLFKFFKDLLGKFLGNN
>
(DSN)A(N2C)(MVA)(DSN)A(NCY)(MVA)
"""

# https://biopython.org/docs/1.81/api/Bio.SeqIO.PirIO.html
PIR1 = """\
>P1;S27231
rhodopsin - northern leopard frog
MNGTEGPNFY IPMSNKTGVV RSPFDYPQYY LAEPWKYSVL AAYMFLLILL GLPINFMTLY
VTIQHKKLRT PLNYILLNLG VCNHFMVLCG FTITMYTSLH GYFVFGQTGC YFEGFFATLG
GEIALWSLVV LAIERYIVVC KPMSNFRFGE NHAMMGVAFT WIMALACAVP PLFGWSRYIP
EGMQCSCGVD YYTLKPEVNN ESFVIYMFVV HFLIPLIIIS FCYGRLVCTV KEAAAQQQES
ATTQKAEKEV TRMVIIMVIF FLICWVPYAY VAFYIFTHQG SEFGPIFMTV PAFFAKSSAI
YNPVIYIMLN KQFRNCMITT LCCGKNPFGD DDASSAATSK TEATSVSTSQ VSPA*

>P1;I51200
rhodopsin - African clawed frog
MNGTEGPNFY VPMSNKTGVV RSPFDYPQYY LAEPWQYSAL AAYMFLLILL GLPINFMTLF
VTIQHKKLRT PLNYILLNLV FANHFMVLCG FTVTMYTSMH GYFIFGPTGC YIEGFFATLG
GEVALWSLVV LAVERYIVVC KPMANFRFGE NHAIMGVAFT WIMALSCAAP PLFGWSRYIP
EGMQCSCGVD YYTLKPEVNN ESFVIYMFIV HFTIPLIVIF FCYGRLLCTV KEAAAQQQES
LTTQKAEKEV TRMVVIMVVF FLICWVPYAY VAFYIFTHQG SNFGPVFMTV PAFFAKSSAI
YNPVIYIVLN KQFRNCLITT LCCGKNPFGD EDGSSAATSK TEASSVSSSQ VSPA*

>P1;JN0120
rhodopsin - Japanese lamprey
MNGTEGDNFY VPFSNKTGLA RSPYEYPQYY LAEPWKYSAL AAYMFFLILV GFPVNFLTLF
VTVQHKKLRT PLNYILLNLA MANLFMVLFG FTVTMYTSMN GYFVFGPTMC SIEGFFATLG
GEVALWSLVV LAIERYIVIC KPMGNFRFGN THAIMGVAFT WIMALACAAP PLVGWSRYIP
EGMQCSCGPD YYTLNPNFNN ESYVVYMFVV HFLVPFVIIF FCYGRLLCTV KEAAAAQQES
ASTQKAEKEV TRMVVLMVIG FLVCWVPYAS VAFYIFTHQG SDFGATFMTL PAFFAKSSAL
YNPVIYILMN KQFRNCMITT LCCGKNPLGD DE-SGASTSKT EVSSVSTSPV SPA*
"""  # noqa: W291 - trailing whitespace

# https://www.bioinformatics.nl/tools/crab_pir.html
PIR2 = """\
>P1;CRAB_ANAPL
ALPHA CRYSTALLIN B CHAIN (ALPHA(B)-CRYSTALLIN).
  MDITIHNPLI RRPLFSWLAP SRIFDQIFGE HLQESELLPA SPSLSPFLMR 
  SPIFRMPSWL ETGLSEMRLE KDKFSVNLDV KHFSPEELKV KVLGDMVEIH 
  GKHEERQDEH GFIAREFNRK YRIPADVDPL TITSSLSLDG VLTVSAPRKQ 
  SDVPERSIPI TREEKPAIAG AQRK*

>P1;CRAB_BOVIN
ALPHA CRYSTALLIN B CHAIN (ALPHA(B)-CRYSTALLIN).
  MDIAIHHPWI RRPFFPFHSP SRLFDQFFGE HLLESDLFPA STSLSPFYLR 
  PPSFLRAPSW IDTGLSEMRL EKDRFSVNLD VKHFSPEELK VKVLGDVIEV 
  HGKHEERQDE HGFISREFHR KYRIPADVDP LAITSSLSSD GVLTVNGPRK 
  QASGPERTIP ITREEKPAVT AAPKK*
"""  # noqa: W291 - trailing whitespace

@unittest.skipIf(SeqIO is None, "BioPython not installed.")
class TestReadingSeq(unittest.TestCase):
    def test_fasta(self):
        for string in (FASTA1, FASTA2, FASTA3, FASTA4):
            gemmi_seqs = gemmi.read_pir_or_fasta(string)
            biopy_seqs = list(SeqIO.parse(io.StringIO(string), 'fasta'))
            self.assertEqual(len(gemmi_seqs), len(biopy_seqs))
            for (gseq, bseq) in zip(gemmi_seqs, biopy_seqs):
                self.assertEqual(gseq.header, bseq.description)
                self.assertEqual(gseq.seq, bseq.seq.rstrip('*'))

    def test_pir(self):
        for string in (PIR1, PIR2):
            gemmi_seqs = gemmi.read_pir_or_fasta(string)
            biopy_seqs = list(SeqIO.parse(io.StringIO(string), 'pir'))
            self.assertEqual(len(gemmi_seqs), len(biopy_seqs))
            for (gseq, bseq) in zip(gemmi_seqs, biopy_seqs):
                g_first, g_desc = gseq.header.splitlines()
                g_type, _, g_id = g_first.partition(';')
                self.assertEqual(g_type, bseq.annotations["PIR-type"])
                self.assertEqual(g_id, bseq.id)
                self.assertEqual(g_desc, bseq.description)
                self.assertEqual(gseq.seq, bseq.seq)

    def test_code_conversion_aa(self):
        seq1 = gemmi.read_pir_or_fasta(FASTA1)[0].seq
        seq3 = gemmi.expand_one_letter_sequence(seq1, gemmi.ResidueKind.AA)
        self.assertEqual(seq1, gemmi.one_letter_code(seq3))

    def test_code_conversion_dna(self):
        seq1 = gemmi.read_pir_or_fasta(FASTA3)[0].seq
        kind = gemmi.ResidueKind.DNA
        seq3 = gemmi.expand_one_letter_sequence(seq1, kind)
        self.assertEqual(seq1, gemmi.one_letter_code(seq3))
        self.assertEqual(seq1, gemmi.pdbx_one_letter_code(seq3, kind))

    def test_code_conversion_rna(self):
        seq1 = 'GGCGAUACCAGCCGAAAGGCCCUUGGCAGCGCC'  # from 8d2b
        kind = gemmi.ResidueKind.RNA
        seq3 = gemmi.expand_one_letter_sequence(seq1, kind)
        self.assertEqual(seq1, gemmi.one_letter_code(seq3))
        self.assertEqual(seq1, gemmi.pdbx_one_letter_code(seq3, kind))

    def test_code_with_brackets(self):
        # test 1PFE _entity_poly.pdbx_seq_one_letter_code[_can]
        seq1 = gemmi.read_pir_or_fasta(FASTA4)[1].seq
        kind = gemmi.ResidueKind.AA
        seq3 = gemmi.expand_one_letter_sequence(seq1, kind)
        self.assertEqual(seq3, ['DSN', 'ALA', 'N2C', 'MVA',
                                'DSN', 'ALA', 'NCY', 'MVA'])
        self.assertEqual(seq1, gemmi.pdbx_one_letter_code(seq3, kind))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

import sys
import gemmi

CCD_PATH = 'components.cif.gz'
OUTPUT_PATH = 'sub_ccd.cif'
COORDINATE_FILES = sys.argv[1:]

# Make a list of residue names that we need.
mon_names = set()
for coordinate_file in COORDINATE_FILES:
    st = gemmi.read_structure(coordinate_file)
    mon_names.update(st[0].get_all_residue_names())

# Copy blocks corresponding to these residues to a new file.
sub = gemmi.cif.Document()
for block in gemmi.cif.read(CCD_PATH):
    if block.name in mon_names:
        sub.add_copied_block(block)
sub.write_file(OUTPUT_PATH)



#!/usr/bin/env python

import sys
import gemmi
import numpy as np
import statsmodels.api as sm

# Usage example: qq.py path/to/df.map 1.5  >plot.dat
assert len(sys.argv) == 3
df_map_path = sys.argv[1]
d_min = float(sys.argv[2])

# read map
df = gemmi.read_ccp4_map(df_map_path, setup=True)
grid = df.grid

# get 1D array of grid points corresponding to a single ASU
map_values = grid.array[grid.masked_asu().mask_array == 0]

# resample data to get as many points as we'd have at the Shannon limit
current_voxel_vol = grid.unit_cell.volume / grid.point_count
shannon_voxel_vol = (d_min / 2)**3
npoints = int(round(current_voxel_vol / shannon_voxel_vol * len(map_values)))

# get quantiles
map_values.sort()
quantile_levels = np.linspace(0, 1, npoints)
data = np.quantile(map_values, quantile_levels)

# print data for the plot
pplot = sm.ProbPlot(data, fit=True)
for x in zip(pplot.theoretical_quantiles,
             pplot.sample_quantiles - pplot.theoretical_quantiles):
    print('%g %g' % x)



#!/usr/bin/env python

import gc
import os
import pickle
import unittest
from gemmi import cif

class TestDoc(unittest.TestCase):
    def test_slicing_and_pickling(self):
        doc = cif.read_string("""
            data_a
            _one 1 _two 2 _three 3
            data_b
            _four 4
            data_c
            _two 2 _four 4 _six 6
        """)
        self.assertTrue('a' in doc)
        self.assertFalse('d' in doc)
        self.assertEqual([b.name for b in doc[:1]], ['a'])
        self.assertEqual([b.name for b in doc[1:]], ['b', 'c'])
        self.assertEqual([b.name for b in doc[:]], ['a', 'b', 'c'])
        self.assertEqual([b.name for b in doc[1:-1]], ['b'])
        self.assertEqual([b.name for b in doc[1:1]], [])

        unpickled = pickle.loads(pickle.dumps(doc))
        self.assertEqual(unpickled.as_string(), doc.as_string())

class TestBlock(unittest.TestCase):
    def test_find(self):
        block = cif.read_string(b"""
            data_test
            _one 1 _two 2 _three 3
            _nonloop_a alpha
            _nonloop_b beta
            loop_ _la _lb _ln A B 1 C D 2
        """).sole_block()
        gc.collect()
        values = block.find_values('_la')
        self.assertEqual(list(values), ['A', 'C'])

        rows = list(block.find(['_lb', '_la']))  # changed order
        gc.collect()
        self.assertEqual(list(rows[0]), ['B', 'A'])
        self.assertEqual(rows[1][1], 'C')

        self.assertEqual(block.get_index('_nonloop_b'), 4)
        self.assertIsNone(block[4].loop)
        self.assertEqual(block[4].pair[0], '_nonloop_b')
        self.assertEqual(block.get_index('_lb'), 5)
        self.assertEqual(block[5].loop.tags[1], '_lb')
        self.assertIsNone(block[5].pair)

        rows = list(block.find('_nonloop_', ['a', 'b']))
        gc.collect()
        self.assertEqual([list(r) for r in rows], [['alpha', 'beta']])

        tab = block.find('_l', ['a', 'n'])
        self.assertEqual(tab.find_row('A')[1], '1')
        self.assertRaises(RuntimeError, tab.find_row, 'B')
        self.assertEqual(tab.find_row('C')[1], '2')
        self.assertEqual(tab.column(0)[1], 'C')
        self.assertEqual(tab.column(1)[0], '1')
        self.assertEqual(tab.find_column('a')[1], 'C')
        self.assertEqual(tab.find_column('_ln')[0], '1')

        tab = block.find(['_lb', '_ln'])
        self.assertRaises(RuntimeError, tab.find_row, 'A')
        self.assertEqual(tab.find_row('B')[1], '1')
        self.assertEqual(tab.find_row('D')[1], '2')

        tab = block.find(['_la', '?_nop', '?_ln'])
        self.assertEqual(len(tab), 2)
        self.assertEqual(tab.width(), 3)
        row = tab[0]
        self.assertIsNotNone(row.get(0))
        self.assertEqual(row[0], 'A')
        self.assertIsNone(row.get(1))
        self.assertEqual(row[2], '1')
        self.assertTrue("None" in repr(row))

    def test_find_values(self):
        v1 = cif.read_string('data_a _v1 one')[0].find_values('_v1')
        gc.collect()
        self.assertListEqual(list(v1), ['one'])
        v2 = cif.read_string('data_a loop_ _v2 a b')[0].find_values('_v2')
        gc.collect()
        self.assertListEqual(list(v2), ['a', 'b'])

    def test_set_pair(self):
        doc = cif.read_string('data_a _a 1 _b 2 _c 3')
        block = doc[0]
        block.set_pair('_d', '9')
        block.set_pair('_b', '8')
        self.assertEqual(block.find_pair('_a'), ('_a', '1'))
        self.assertEqual(block.find_value('_b'), '8')
        self.assertEqual(block.find_value('_c'), '3')
        self.assertEqual(block.find_value('_d'), '9')

    def test_set_pairs(self):
        doc = cif.read_string('data_a _zza 0 _z_a 1 _Z_b 2 _z_C 3 _zzb 4')
        block = doc[0]
        expected = [('_zza', '0'), ('_z_a', '1'), ('_Z_b', '2'),
                    ('_z_C', '3'), ('_zzb', '4')]
        self.assertEqual([v.pair for v in block], expected)
        block.set_pairs('_z_', {'1': 5})
        expected.insert(4, ('_z_1', '5'))
        self.assertEqual([v.pair for v in block], expected)
        block.set_pairs('_z_', {'B': 10, 'c' : 11, 'a': 9})
        expected[1:4] = [('_z_a', '9'), ('_z_B', '10'), ('_z_c', '11')]
        self.assertEqual([v.pair for v in block], expected)

    def test_set_loop(self):
        block = cif.read_string('data_a _c.a 1 _c.b 2 _c.c 3 loop_ _cx.b 3')[0]
        block.init_loop('_cx.', ['b']).add_row(['x'])
        self.assertEqual(block.find_value('_c.a'), '1')
        block.init_mmcif_loop('_c.', ['c']).add_row(['y'])
        self.assertEqual(block.find_value('_c.a'), None)
        loop = block.init_loop('_c.', ['c', 'd'])
        loop.set_all_values([list('one'), list('two')])
        self.assertEqual(loop.width(), 2)
        self.assertEqual(loop.length(), 3)
        self.assertEqual(list(block.find(['_c.c', '_c.d'])[1]), ['n', 'w'])
        tab = block.find_or_add('_c.', ['d'])
        self.assertEqual((tab.width(), len(tab)), (1, 3))
        tab = block.find_or_add('_c.', ['d', 'c'])
        self.assertEqual((tab.width(), len(tab)), (2, 3))
        tab = block.find_or_add('_u.', ['d', 'c'])
        self.assertEqual((tab.width(), len(tab)), (2, 0))

    def test_erase(self):
        block = cif.read_string('data_a _a_a 1 _a_b 2 _a_c 3 loop_ _x 4')[0]
        block.find('_a_', ['a', '?x', '?c', '?bb']).erase()
        self.assertEqual(block.as_string().split(),
                         ['data_a', '_a_b', '2', 'loop_', '_x', '4'])

    def test_setitem(self):
        block = cif.read_string('data_a _a 1 _b 2 _c 3')[0]
        self.assertEqual(block.find_value('_b'), '2')
        col_b = block.find_values('_b')
        col_b[0] = '20'
        self.assertEqual(block.find_value('_b'), '20')
        bc = block.find(['_b', '_a'])
        self.assertEqual(bc[0]['_a'], '1')
        bc[0][0] = '30'
        self.assertEqual(block.find_value('_b'), '30')
        bc[0][1] = '40'
        self.assertEqual(block.find_value('_a'), '40')
        self.assertEqual(bc[0]['_a'], '40')
        bc[0]['_a'] = '44'
        self.assertEqual(block.find_value('_a'), '44')
        self.assertEqual(block.find_value('_a'), '44')
        self.assertEqual(block.find_value('_b'), '30')

    def test_add_row(self):
        block = cif.read_string('data_a loop_ _x _y 1 2 3 4')[0]
        loop = block.find_loop('_x').get_loop()
        loop.add_row(['5', '6'])
        self.assertEqual(loop.values, ['1', '2', '3', '4', '5', '6'])
        loop.add_row(['?', '0'], 0)
        self.assertEqual(list(block.find_values('_y')), '0 2 4 6'.split())
        self.assertEqual(loop.length(), 4)
        block.find(['_x']).append_row(['xa'])
        block.find(['_y']).append_row(['ya'])
        block.find(['_y', '_x']).append_row(['yb', 'xb'])
        loop.add_columns(column_names=['_a', '_z'], value='A')
        loop.remove_column('_a')
        self.assertEqual(loop.tags, ['_x', '_y', '_z'])
        block.find(['_x', '_y']).append_row(['xc', 'yc'])
        self.assertEqual(loop.length(), 8)
        self.assertEqual(list(block.find_values('_x')),
                         '? 1 3 5 xa . xb xc'.split())
        self.assertEqual(list(block.find_values('_y')),
                         '0 2 4 6 . ya yb yc'.split())
        self.assertEqual(list(block.find_values('_z')),
                         'A A A A A A A .'.split())

    def test_set_mmcif_category(self):
        doc = cif.Document()
        block = doc.add_new_block('b')
        block.set_mmcif_category('_c', {
            'one': ['?', 'ab', ';text field\n;'],
            'two': [-1, 4./3, '"double quoted"']}, raw=True)
        self.assertEqual(block.find_values('_c.one')[0], '?')
        self.assertEqual(block.find_values('_c.one').str(1), 'ab')
        self.assertEqual(block.find_values('_c.one').str(2), 'text field')
        self.assertEqual(block.find_values('_c.two').str(0), '-1')
        self.assertEqual(block.find_values('_c.two').str(2), 'double quoted')
        block.set_mmcif_category('_d', {
            'one': [None, 'a b', 'text\nfield'],
            'two': [-1, '?', False]})
        def check_d():
            self.assertEqual(block.find_values('_d.one')[0], '?')
            self.assertEqual(block.find_values('_d.one').str(1), 'a b')
            self.assertEqual(block.find_values('_d.one').str(2), 'text\nfield')
            self.assertEqual(block.find_values('_d.one')[2], ';text\nfield\n;')
            self.assertEqual(block.find_values('_d.two').str(0), '-1')
            self.assertEqual(block.find_values('_d.two').str(1), '?')
            self.assertEqual(block.find_values('_d.two')[2], '.')
        check_d()
        block.set_mmcif_category('_D', {
            'one': ['?', "'a b'", ';text\nfield\n;'],
            'two': ['-1', "'?'", '.']},
            raw=True)
        self.assertEqual(set(block.find_mmcif_category('_d').tags),
                         {'_D.one', '_D.two'})
        check_d()
        block.set_mmcif_category('_d', {
            'one': [None, "'a b'", ';text\nfield\n;'],
            'two': [-1, "'?'", False]},
            raw=True)
        check_d()
        block.set_mmcif_category('_d', block.get_mmcif_category('_D'))
        check_d()
        block.set_mmcif_category('_d', block.get_mmcif_category('_d', raw=True),
                                 raw=True)
        check_d()
        block.set_mmcif_category('_d', {})
        self.assertEqual(block.find_mmcif_category('_d').width(), 0)

    def test_mmcif_file(self):
        path = os.path.join(os.path.dirname(__file__), '5i55.cif')
        block = cif.read(path).sole_block()
        self.assertEqual(len(block.get_mmcif_category_names()), 54)
        entry_cat = block.get_mmcif_category('_entry')
        self.assertEqual(entry_cat, {'id': ['5I55']})
        drw_cat = block.get_mmcif_category('_diffrn_radiation_wavelength.')
        self.assertEqual(drw_cat, {'id': ['1', '2', '3'],
                                   'wavelength': ['0.9792', '0.9794', '0.9796'],
                                   'wt': ['1.0']*3})
        cc_cat = block.get_mmcif_category('_chem_comp.')
        self.assertEqual(cc_cat['mon_nstd_flag'][:2], [False, 'y'])
        self.assertEqual(cc_cat['pdbx_synonyms'][:2], [None, None])

        # test __delitem__
        del block.find(['_entry.id'])[0]
        entry_cat = block.get_mmcif_category('_entry')
        self.assertEqual(entry_cat, {'id': []})
        def nums():
            return list(block.find_values('_entity_poly_seq.num'))
        tab = block.find(['_entity_poly_seq.mon_id'])
        self.assertEqual(nums(), [str(i) for i in range(1, 23)])
        del tab[1::2]
        self.assertEqual(nums(), [str(i) for i in range(1, 23, 2)])
        del tab[3:]
        self.assertEqual(nums(), ['1', '3', '5'])
        del tab[:-1]
        self.assertEqual(nums(), ['5'])

    def test_reading_gzipped_file(self):
        path = os.path.join(os.path.dirname(__file__), '1pfe.cif.gz')
        cif_doc = cif.read(path)
        block = cif_doc.sole_block()
        categories = block.get_mmcif_category_names()
        self.assertEqual(categories[0], '_entry.')
        self.assertEqual(len(categories), 72)
        exptl = block.find_mmcif_category('_exptl')
        self.assertEqual(exptl.get_prefix(), '_exptl.')
        self.assertEqual(list(exptl.tags), ['_exptl.entry_id', '_exptl.method',
                                            '_exptl.crystals_number'])
        self.assertEqual(len(exptl), 1)
        self.assertEqual(exptl.width(), 3)
        exptl = block.find_mmcif_category('_exptl')
        self.assertEqual(len(exptl), 1)
        self.assertEqual(exptl.width(), 3)
        self.assertEqual(exptl[0].str(1), 'X-RAY DIFFRACTION')
        struct_asym = block.find_mmcif_category('_struct_asym')
        self.assertEqual(len(struct_asym), 7)
        self.assertEqual(struct_asym.width(), 5)
        self.assertListEqual(list(struct_asym[3]), ['D', 'N', 'N', '4', '?'])
        nonexistent = block.find_mmcif_category('_nonexistent')
        self.assertRaises(RuntimeError, nonexistent.get_prefix)
        self.assertEqual(len(nonexistent), 0)
        self.assertEqual(nonexistent.width(), 0)

    def test_file_not_found(self):
        with self.assertRaises(IOError):
            cif.read('file-that-does-not-exist.cif')

    def test_syntax_error(self):
        with self.assertRaises(ValueError):
            cif.read_string('data_a boom')

    def test_line_endings(self):
        lines = ['data_a', '_a_field', ';', 'text line', '2nd line', ';']
        for eol in ('\r\n', '\n'):
            input_str = eol.join(lines) + eol
            doc = cif.read_string(input_str)
            output_str = doc.as_string()
            self.assertEqual(input_str.replace(eol, '\n'), output_str)

    def test_text_field_eol(self):
        # use fragment of ma-bak-cepc-0017.cif
        path = os.path.join(os.path.dirname(__file__), 'eol-test.cif')
        with open(path) as f:
            cif_string = f.read()
        doc = cif.read_string(cif_string)
        output = doc.as_string()
        self.assertEqual([s.rstrip() for s in cif_string.splitlines()],
                         output.splitlines())

    def test_write_style(self):
        doc = cif.read_string('data_one _x y')
        self.assertEqual(doc.as_string(), doc[0].as_string())
        self.assertEqual(doc.as_string().splitlines(), ['data_one', '_x y'])
        options = cif.WriteOptions()
        options.align_pairs = 33
        self.assertEqual(doc.as_string(options).splitlines(),
                         ['data_one', '_x                                y'])

    def test_relion_syntax_exception(self):
        block = cif.read_string("""\
            data_
            loop_
            _rlnImageName
            _rlnMicrographName
            _rlnDefocusU
            _rlnDefocusV
            mic1/img000001.spi mic1 10000 10500
            mic1/img000002.spi mic1 10000 10501
            """).sole_block()
        self.assertEqual(block.name, ' ')
        item = block[0]
        self.assertEqual(item.line_number, 2)
        table = block.item_as_table(item)
        self.assertEqual(table[1][3], '10501')

    def test_case_sensitivity(self):
        block = cif.read_string("""
            daTA_test
            _One 1 _two 2 _thrEE 3
            _NonLoop_a alpha
            _nonloop_B beta
            loop_ _laAa _lbBb _ln A B 1 C D 2
        """).sole_block()
        values = block.find_values('_laaA')
        self.assertEqual(list(values), ['A', 'C'])

        rows = list(block.find(['_lBbb', '_lAaa']))  # changed order
        self.assertEqual(list(rows[0]), ['B', 'A'])

        self.assertEqual(block.get_index('_nonlOOp_b'), 4)
        self.assertEqual(block.get_index('_lBBB'), 5)
        self.assertEqual(block.find_pair('_Three'), ('_thrEE', '3'))

        values.erase()
        block.find_values('_two').erase()
        block.find_values('_nonloop_b').erase()
        expected = """\
            data_test
            _One 1 _thrEE 3
            _NonLoop_a alpha
            loop_ _lbBb _ln  B 1  D 2"""
        block_str = block.as_string()
        self.assertEqual(block_str.split(), expected.split())

        unpickled = pickle.loads(pickle.dumps(block))
        self.assertEqual(unpickled.as_string(), block_str)

class TestQuote(unittest.TestCase):
    def test_quote(self):
        self.assertEqual(cif.quote('a.b-c'), 'a.b-c')
        self.assertEqual(cif.quote('loop_'), "'loop_'")
        self.assertEqual(cif.quote('a"\nb'), ';a"\nb\n;')
        self.assertEqual(cif.quote(u'\u0394'), u"'\u0394'")

def full_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)

class TestDictionary(unittest.TestCase):
    def test_frame_reading(self):
        block = cif.read(full_path('mmcif_pdbx_v50_frag.dic')).sole_block()
        self.assertIsNone(block.find_frame('heyho'))
        frame = block.find_frame('_atom_site.auth_atom_id')
        code = frame.find_value('_item_type.code')
        self.assertEqual(code, 'atcode')

    def test_validation(self):
        doc = cif.read_string("""
            data_dummy_block
            loop_
            _atom_site.auth_asym_id
            _atom_site.auth_comp_id
            _atom_site.auth_atom_id
            A CYS N
            A CYS CA
            _custom_tag 5
            loop_ _another_one 6
        """)
        msg_list = []
        ddl = cif.Ddl(logger=(lambda msg: msg_list.append(msg), 6))
        ddl.read_ddl(cif.read(full_path('mmcif_pdbx_v50_frag.dic')))
        self.assertEqual(msg_list, [])
        ddl.validate_cif(doc)
        self.assertEqual(msg_list,
                         ['[dummy_block] unknown tag _custom_tag',
                          '[dummy_block] unknown tag _another_one',
                          '[dummy_block] missing mandatory tag: _atom_site.id',
                          '[dummy_block] missing category key: _atom_site.id'])

        # try out set_logger()
        counter = 0
        def incr_counter(_):
            nonlocal counter
            counter += 1
        ddl.set_logger((incr_counter, 7))
        ddl.validate_cif(doc)
        self.assertEqual(counter, len(msg_list))

        msg_list = []
        ddl.set_logger(lambda msg: msg_list.append(msg))
        doc = cif.read_string("""
            data_b2
            _something.unexpected here
            loop_
            _atom_site.id
            _atom_site.auth_asym_id
            _atom_site.auth_comp_id
            _atom_site.auth_atom_id
            _atom_site.attached_hydrogens
            1 A CYS N   -1
            2 'hey hey' CYS CA   1
            2 A CYS CA   1

            _exptl_crystal.id 1
            _exptl_crystal.density_percent_sol   10.5
        """)
        ddl.validate_cif(doc)
        expected = [
            '[b2] unknown tag _something.unexpected',                 # 1
            "string:4 [b2] _atom_site.auth_asym_id: "
            "'hey hey' does not match the code regex",                # 2
            'string:4 [b2] _atom_site.attached_hydrogens: '
            'value out of expected range: -1',                        # 3
            '[b2] category atom_site has 1 duplicated key:\n  id=2',  # 4
        ]
        self.assertEqual(msg_list, expected)
        msg_list = []
        ddl.use_deposition_checks = True
        ddl.validate_cif(doc)
        expected.insert(-1, 'string:15 [b2] value out of expected range: 10.5')
        self.assertEqual(msg_list, expected)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# This script looks for chains that exceed the size of the unit cell (by >20%)
# in one of the a, b, c directions.

import sys
import gemmi

def run(path):
    counter = 0
    st = gemmi.read_structure(path)
    if st.cell.is_crystal():
        st.add_entity_types()
        for chain in st[0]:
            polymer = chain.get_polymer()
            if polymer:
                low_bounds = [float('+inf')] * 3
                high_bounds = [float('-inf')] * 3
                for residue in polymer:
                    for atom in residue:
                        pos = st.cell.fractionalize(atom.pos)
                        for i in range(3):
                            if pos[i] < low_bounds[i]:
                                low_bounds[i] = pos[i]
                            if pos[i] > high_bounds[i]:
                                high_bounds[i] = pos[i]
                for i in range(3):
                    delta = high_bounds[i] - low_bounds[i]
                    if delta > 1.2:  # 120% of the unit cell size
                        counter += 1
                        code = st.info['_entry.id']
                        print('%s   chain:%s   delta%c = %.3f' %
                              (code, chain.name, ord('X') + i, delta))
    return counter

def main():
    for arg in sys.argv[1:]:
        for path in gemmi.CoorFileWalk(arg):
            run(path)

if __name__ == '__main__':
    main()



import gemmi
from rdkit.Chem import AllChem as rd

BOND_TYPE_TO_RDKIT = {
    gemmi.BondType.Unspec:   rd.BondType.UNSPECIFIED,
    gemmi.BondType.Single:   rd.BondType.SINGLE,
    gemmi.BondType.Double:   rd.BondType.DOUBLE,
    gemmi.BondType.Triple:   rd.BondType.TRIPLE,
    gemmi.BondType.Aromatic: rd.BondType.AROMATIC,
    gemmi.BondType.Deloc:    rd.BondType.OTHER,
    gemmi.BondType.Metal:    rd.BondType.OTHER,
}

def chemcomp_to_rdkit(cc: gemmi.ChemComp):
    em = rd.EditableMol(rd.Mol())
    for atom in cc.atoms:
        a = rd.Atom(atom.el.atomic_number)
        a.SetMonomerInfo(rd.AtomPDBResidueInfo(atom.id))
        em.AddAtom(a)
    idx = {atom.id: n for (n, atom) in enumerate(cc.atoms)}
    for bond in cc.rt.bonds:
        order = BOND_TYPE_TO_RDKIT[bond.type]
        if bond.aromatic:
            order = rd.BondType.AROMATIC
        em.AddBond(idx[bond.id1.atom], idx[bond.id2.atom], order=order)
    return em.GetMol()


if __name__ == '__main__':
    import sys
    path = sys.argv[1]
    block = gemmi.cif.read(path)[-1]
    cc = gemmi.make_chemcomp_from_block(block)
    m = chemcomp_to_rdkit(cc)
    print(rd.MolToSmiles(m))



#!/usr/bin/env python
# Check column presence and order in the _atom_site category.
# In mmCIF v5 esd _atom_site.*esd columns were removed.

import sys
from gemmi import cif, CifWalk
from typing import Dict

#ESD = 'Cartn_x_esd Cartn_y_esd Cartn_z_esd occupancy_esd B_iso_or_equiv_esd '
ESD = ''
USUAL_ORDER = ('group_PDB id type_symbol label_atom_id label_alt_id '
               'label_comp_id label_asym_id label_entity_id label_seq_id '
               'pdbx_PDB_ins_code Cartn_x Cartn_y Cartn_z occupancy '
               'B_iso_or_equiv ' + ESD + 'pdbx_formal_charge '
               'auth_seq_id auth_comp_id auth_asym_id auth_atom_id '
               'pdbx_PDB_model_num')
counts : Dict[str, int] = {}
for arg in sys.argv[1:]:
    for path in CifWalk(arg):
        block = cif.read(path).sole_block()
        loop_tags = block.find_loop("_atom_site.id").get_loop().tags
        assert all(t.startswith("_atom_site.") for t in loop_tags)
        tags = ' '.join(t[11:] for t in loop_tags)
        if tags != USUAL_ORDER:
            print(tags)
            print(USUAL_ORDER)
            print(block.name, tags)
            counts[tags] = counts.get(tags, 0) + 1

for key, value in counts.items():
    print(value, key)

# Results: in v4 a few EM structures (5A9Z 5AA0 5FKI 4UDF)
# had different order, with ATOM/HETATM in the middle.



#!/usr/bin/env python3

# List CCD entries that contain the specified entry as a substructure.
# Ignoring hydrogens and bond types.

import sys
import networkx
from networkx.algorithms import isomorphism
import gemmi

CCD_PATH = 'components.cif.gz'

def graph_from_block(block):
    cc = gemmi.make_chemcomp_from_block(block)
    cc.remove_hydrogens()
    G = networkx.Graph()
    for atom in cc.atoms:
        G.add_node(atom.id, Z=atom.el.atomic_number)
    for bond in cc.rt.bonds:
        G.add_edge(bond.id1.atom, bond.id2.atom)
    return G

def main():
    assert len(sys.argv) == 2, "Usage: ccd_subgraph.py three-letter-code"
    ccd = gemmi.cif.read(CCD_PATH)
    entry = sys.argv[1]
    pattern = graph_from_block(ccd[entry])
    pattern_nodes = networkx.number_of_nodes(pattern)
    pattern_edges = networkx.number_of_edges(pattern)
    node_match = isomorphism.categorical_node_match('Z', 0)
    for block in ccd:
        G = graph_from_block(block)
        GM = isomorphism.GraphMatcher(G, pattern, node_match=node_match)
        if GM.subgraph_is_isomorphic():
            print(block.name, '\t +%d nodes, +%d edges' % (
                networkx.number_of_nodes(G) - pattern_nodes,
                networkx.number_of_edges(G) - pattern_edges))

main()



#!/usr/bin/env python
#
# This script reads monomer definitions from a gzipped CIF file (components.cif.gz)
# and extracts specified monomer blocks into separate CIF files.
#
# For example, to extract the 'ARG' and 'ATP' monomers:
#   python use.py ARG ATP
# This will create two files: ./orig/ARG.cif and ./orig/ATP.cif.
#
# To extract all monomers starting with 'A':
#   python use.py 'A*'
# This will create files in ./orig/a/ directory.

import sys
import os
import gzip
import subprocess
import fnmatch

# Programmatically add the gemmi build path to sys.path
# This allows 'import gemmi' to succeed without an explicit PYTHONPATH env var.
gemmi_build_path = os.path.join(os.path.expanduser("~"), "gemmi", "gemmi", "build", "py")
if gemmi_build_path not in sys.path:
    sys.path.insert(0, gemmi_build_path)

try:
    import gemmi
except ImportError:
    print("Error: The 'gemmi' module could not be imported.")
    print("Please ensure that the 'gemmi' library is installed at:")
    print(f"  {gemmi_build_path}")
    print("Or that it's otherwise available in your Python environment.")
    sys.exit(1)


def run_acedrg(code: str, input_file: str, output_dir: str):
    """
    Runs the Acedrg command for a given monomer.

    Args:
        code: The monomer code (e.g., 'ATP').
        input_file: Path to the input CIF file (e.g., 'orig/ATP.cif').
        output_dir: The directory for Acedrg output (e.g., 'acedrg').
    """
    output_cif = os.path.join(output_dir, code + ".cif")
    if os.path.exists(output_cif):
        print(f"Skipping {code}: {output_cif} already exists")
        return

    if not os.path.exists(output_dir):
        try:
            os.makedirs(output_dir)
            print(f"Created directory: {output_dir}")
        except OSError as e:
            print(f"Error creating directory {output_dir}: {e}")
            return  # Stop if we can't create the output directory

    output_prefix = os.path.join(output_dir, code)
    command = ["acedrg", "-c", input_file, "-o", output_prefix]
    print(f"\nRunning: {' '.join(command)}")

    try:
        # Execute the command, capturing output and checking for errors
        _ = subprocess.run(
            command, check=True, capture_output=True, text=True, encoding='utf-8'
        )
        print(f"Acedrg finished successfully for {code}.")
        # Optionally, uncomment the following lines to see Acedrg's output
        # if _.stdout:
        #     print("Acedrg Log:\n" + _.stdout)
        # if _.stderr:
        #     print("Acedrg Errors:\n" + _.stderr)

    except FileNotFoundError:
        print("\nError: 'acedrg' command not found.")
        print("Please ensure Acedrg is installed and its location is in your system's PATH.")
        # Exit the script because Acedrg is a required part of the process
        sys.exit(1)
    except subprocess.CalledProcessError as e:
        print(f"\nError running Acedrg for {code}:")
        print(f"Command failed: {' '.join(command)}")
        print(f"Return code: {e.returncode}")
        print(f"--- Acedrg STDOUT ---\n{e.stdout}")
        print(f"--- Acedrg STDERR ---\n{e.stderr}")
    except Exception as e:
        print(f"\nAn unexpected error occurred while running Acedrg for {code}: {e}")


def extract_and_process_monomers(
    monomer_codes: list[str], input_cif: str, orig_dir: str, acedrg_dir: str
):
    """
    Reads a CIF file, finds monomer blocks, writes them to files, and runs Acedrg.

    Args:
        monomer_codes: A list of monomer codes to extract.
        input_cif: Path to the input gzipped CIF file.
        orig_dir: The directory to save the initial monomer .cif files.
        acedrg_dir: The directory for the Acedrg output.
    """
    if not os.path.exists(orig_dir):
        try:
            os.makedirs(orig_dir)
            print(f"Created directory: {orig_dir}")
        except OSError as e:
            print(f"Error creating directory {orig_dir}: {e}")
            sys.exit(1)

    try:
        with gzip.open(input_cif, "rt", encoding="utf-8") as gz_file:
            cif_content = gz_file.read()
        doc = gemmi.cif.read_string(cif_content)
    except FileNotFoundError:
        print(f"Error: Input file not found at '{input_cif}'")
        sys.exit(1)
    except Exception as e:
        print(f"Error reading or parsing '{input_cif}': {e}")
        sys.exit(1)

    for code in monomer_codes:
        block = doc.find_block(code)
        if block:
            new_doc = gemmi.cif.Document()
            new_doc.add_copied_block(block)

            output_filename = f"{code}.cif"
            output_path = os.path.join(orig_dir, output_filename)

            try:
                new_doc.write_file(output_path)
                print(f"Successfully wrote {output_path}")
                # After writing the file, run Acedrg
                run_acedrg(code, output_path, acedrg_dir)
            except Exception as e:
                print(f"Error writing file or running Acedrg for monomer '{code}': {e}")
        else:
            print(f"\nWarning: Monomer '{code}' not found in '{input_cif}'")


def load_cif_document(input_cif: str):
    """Load and parse a gzipped CIF file, returning the document."""
    try:
        with gzip.open(input_cif, "rt", encoding="utf-8") as gz_file:
            cif_content = gz_file.read()
        return gemmi.cif.read_string(cif_content)
    except FileNotFoundError:
        print(f"Error: Input file not found at '{input_cif}'")
        sys.exit(1)
    except Exception as e:
        print(f"Error reading or parsing '{input_cif}': {e}")
        sys.exit(1)


def find_matching_blocks(doc, pattern: str) -> list[str]:
    """Find all block names matching a wildcard pattern."""
    pattern_upper = pattern.upper()
    matching = []
    for block in doc:
        if fnmatch.fnmatch(block.name, pattern_upper):
            matching.append(block.name)
    return matching


def main():
    """
    Main function to parse command-line arguments and initiate extraction and processing.
    """
    if len(sys.argv) < 2:
        print(f"Usage: python {sys.argv[0]} MONOMER1 MONOMER2 ...")
        print(f"       python {sys.argv[0]} 'A*'  # extract all A* to orig/a/")
        print(f"Example: python {sys.argv[0]} ARG ATP")
        sys.exit(1)

    input_cif_file = "components.cif.gz"
    base_orig_dir = "orig"
    base_acedrg_dir = "acedrg"

    # Check if any argument contains a wildcard
    has_wildcard = any('*' in arg or '?' in arg for arg in sys.argv[1:])

    if has_wildcard:
        # Load document once for pattern matching
        print(f"Loading {input_cif_file}...")
        doc = load_cif_document(input_cif_file)

        for pattern in sys.argv[1:]:
            pattern_upper = pattern.upper()
            if '*' in pattern or '?' in pattern:
                # Wildcard pattern - find matching blocks
                matching_codes = find_matching_blocks(doc, pattern_upper)
                if not matching_codes:
                    print(f"Warning: No monomers match pattern '{pattern}'")
                    continue

                # Determine subdirectory from pattern prefix (e.g., 'A*' -> 'a')
                prefix = pattern_upper.split('*')[0].split('?')[0].lower()
                if prefix:
                    orig_dir = os.path.join(base_orig_dir, prefix)
                    acedrg_dir = os.path.join(base_acedrg_dir, prefix)
                else:
                    orig_dir = base_orig_dir
                    acedrg_dir = base_acedrg_dir

                print(f"Pattern '{pattern}' matched {len(matching_codes)} monomers")
                print(f"Output directories: {orig_dir}, {acedrg_dir}")

                # Process each matching monomer
                if not os.path.exists(orig_dir):
                    os.makedirs(orig_dir)
                    print(f"Created directory: {orig_dir}")

                for code in matching_codes:
                    block = doc.find_block(code)
                    if block:
                        new_doc = gemmi.cif.Document()
                        new_doc.add_copied_block(block)
                        output_path = os.path.join(orig_dir, f"{code}.cif")
                        try:
                            new_doc.write_file(output_path)
                            print(f"Wrote {output_path}")
                            run_acedrg(code, output_path, acedrg_dir)
                        except Exception as e:
                            print(f"Error writing {code}: {e}")
            else:
                # Single monomer code
                code = pattern_upper
                extract_and_process_monomers([code], input_cif_file, base_orig_dir, base_acedrg_dir)
    else:
        # No wildcards - use original behavior
        monomer_codes = [arg.upper() for arg in sys.argv[1:]]
        extract_and_process_monomers(monomer_codes, input_cif_file, base_orig_dir, base_acedrg_dir)


if __name__ == "__main__":
    main()




import os
import urllib.request
import urllib.error
import time
from typing import Optional

import gemmi

def get_url_for_code(pdb_id: str, use_cif: bool = True, pdb_site: str = 'R') -> str:
    pdb_site = pdb_site.upper()[0]
    if pdb_site == 'R':
        url_prefix = f"https://files.rcsb.org/download/{pdb_id.upper()}"
        if use_cif:
            return f"{url_prefix}.cif"
        else:
            return f"{url_prefix}.pdb"
    elif pdb_site == 'E':
        url_prefix = "https://www.ebi.ac.uk/pdbe/entry-files/download"
        if use_cif:
            return f"{url_prefix}/{pdb_id.lower()}.cif"
        else:
            return f"{url_prefix}/pdb{pdb_id.lower()}.ent"
    elif pdb_site == 'U': # also 'E', but _updated cif file
        url_prefix = "https://www.ebi.ac.uk/pdbe/entry-files/download"
        if use_cif:
            return f"{url_prefix}/{pdb_id.lower()}_updated.cif"
        else:
            raise RuntimeError(f"updated PDBe files must be in mmCIF not PDB format")
    elif pdb_site == 'J':
        filetype = ('M' if use_cif else 'P')
        return f"https://data.pdbj.org/pub/pdb/data/{gemmi.path_in_pdb_dir(pdb_id, filetype)}"
    else:
        raise RuntimeError(f"{pdb_site=}; expected 'R', 'E','U', or 'J' (RCSB/Europe/E-updated/Japan)")

# fetch and read a file for a PDB entry from RCSB/PDBe/PDBj/PDB_REDO
# if not available locally
def read_structure_with_code(pdb_id: str, use_cif: bool = True, pdb_site: str = 'R') -> Optional[gemmi.Structure]:
    assert gemmi.is_pdb_code(pdb_id)
    local_path = gemmi.expand_if_pdb_code(pdb_id[0])
    if os.path.isfile(local_path):
        return gemmi.read_structure(local_path)
    url = get_url_for_code(pdb_id, use_cif, pdb_site)
    for i in range(3):
        try:
            with urllib.request.urlopen(url, timeout=10) as response:
                content = response.read()
                if use_cif:
                    doc = gemmi.cif.read_string(content)
                    return gemmi.make_structure_from_block(doc[0])
                else:
                    return gemmi.read_pdb_string(content)
        except urllib.error.URLError as e:
            if i == 3:
                raise
            print(f"Retrying {url} after error: {e}")
            time.sleep(2 * (i + 1))
    return None

def main():
    import sys
    import argparse
    parser = argparse.ArgumentParser(description='Fetch PDB/CIF files.')
    parser.add_argument('-p', '--pdb', action='store_true',
                        help='fetch PDB format (default: mmCIF)')
    site_group = parser.add_mutually_exclusive_group()
    site_group.add_argument('-e', '--pdbe', action='store_const', const='E', dest='pdb_site',
                            help='fetch from PDBe')
    site_group.add_argument('-u', '--udpated', action='store_const', const='u', dest='pdb_site',
                            help='fetch "updated" file from PDBe')
    site_group.add_argument('-j', '--pdbj', action='store_const', const='J', dest='pdb_site',
                            help='fetch from PDBj')
    parser.set_defaults(pdb_site='R')
    parser.add_argument('pdb_ids', nargs='+', help='PDB IDs')
    args = parser.parse_args()

    use_cif = not args.pdb

    for pdb_id in args.pdb_ids:
        url = get_url_for_code(pdb_id, use_cif=use_cif, pdb_site=args.pdb_site)
        print(f'get {url}...')
        with urllib.request.urlopen(url, timeout=5) as response:
            content = response.read()
            ext = ('.cif' if use_cif else '.pdb')
            if args.pdb_site == 'u':
                ext = '_updated.cif'
            output_path = pdb_id + ext
            print(f'write {output_path}...')
            with open(output_path, 'wb') as output_file:
                output_file.write(content)

if __name__ == '__main__':
    main()



#!/usr/bin/env python

import os
import unittest
import gemmi

SO2_FROM_MONOMER = """\
CRYST1    1.000    1.000    1.000  90.00  90.00  90.00 P 1                      
HETATM    1  S   SO3     1      -5.979   0.717  17.353  1.00 20.00           S  
HETATM    2  O1  SO3     1      -5.035   1.876  17.325  1.00 20.00           O  
HETATM    3  O2  SO3     1      -7.003   1.053  16.315  1.00 20.00           O1-
HETATM    4  O3  SO3     1      -5.199  -0.407  16.748  1.00 20.00           O1-
"""  # noqa: W291 - trailing whitespace

def full_path(filename):
    return os.path.join(os.path.dirname(__file__), filename)

# tests for make_structure_from_chemcomp_block()
class TestChemCompXyz(unittest.TestCase):
    def test_reading_monomer_SO3_coordinates(self):
        path = full_path('SO3.cif')
        block = gemmi.cif.read(path)[-1]
        st = gemmi.make_structure_from_chemcomp_block(block)
        out = st.make_pdb_string(gemmi.PdbWriteOptions(minimal=True))
        self.assertEqual(out.splitlines(), SO2_FROM_MONOMER.splitlines())

    # comparing HEM.cif from PDB CCD with HEM.pdb from PDBe
    def test_reading_HEM(self):
        cif_path = full_path('HEM.cif')
        cif_block = gemmi.cif.read(cif_path).sole_block()

        cif_st = gemmi.make_structure_from_chemcomp_block(cif_block)
        self.assertEqual(len(cif_st), 2)  # two models: Ideal and Example

        which = gemmi.ChemCompModel.Xyz
        cif_st = gemmi.make_structure_from_chemcomp_block(cif_block, which)
        self.assertEqual(len(cif_st), 0)

        which = gemmi.ChemCompModel.Xyz | gemmi.ChemCompModel.Ideal
        cif_st = gemmi.make_structure_from_chemcomp_block(cif_block, which)
        self.assertEqual(len(cif_st), 1)

        # PDBe files have residue number 0 and ATOM instead of HETATM
        residue = cif_st[0][0][0]
        residue.seqid.num = 0
        residue.het_flag = 'A'
        for atom in residue:
            atom.b_iso = 20
        minimal_opt = gemmi.PdbWriteOptions(minimal=True)
        cif_out = cif_st.make_pdb_string(minimal_opt)
        pdb_path = full_path('HEM.pdb')
        pdb_st = gemmi.read_structure(pdb_path)
        pdb_out = pdb_st.make_pdb_string(minimal_opt)
        self.assertEqual(cif_out.splitlines(), pdb_out.splitlines())

    # HEN.cif from CCD does not provide ideal coordinates
    def test_reading_HEN(self):
        path = full_path('HEN.cif')
        block = gemmi.cif.read(path).sole_block()
        st = gemmi.make_structure_from_chemcomp_block(block)
        self.assertEqual(len(st), 1)

# tests for gemmi/chemcomp.hpp
class TestChemComp(unittest.TestCase):
    def test_hem_ccd(self):
        path = os.path.join(os.path.dirname(__file__), 'HEM.cif')
        block = gemmi.cif.read(path).sole_block()
        cc = gemmi.make_chemcomp_from_block(block)
        self.assertEqual(len(cc.atoms), 75)
        self.assertEqual(len(cc.rt.bonds), 82)
        bond = cc.rt.get_bond("CGA", "O1A")
        self.assertEqual(bond.type, gemmi.BondType.Double)
        self.assertEqual(bond.aromatic, False)
        bond = cc.rt.get_bond("NA", "C1A")
        self.assertEqual(bond.type, gemmi.BondType.Single)
        self.assertEqual(bond.aromatic, True)

    def test_hen_ccd(self):
        path = os.path.join(os.path.dirname(__file__), 'HEN.cif')
        block = gemmi.cif.read(path).sole_block()
        cc = gemmi.make_chemcomp_from_block(block)
        self.assertEqual(len(cc.atoms), 45)
        self.assertEqual(len(cc.rt.bonds), 45)

    def test_so2_ccp4(self):
        path = os.path.join(os.path.dirname(__file__), 'SO3.cif')
        block = gemmi.cif.read(path)[-1]
        cc = gemmi.make_chemcomp_from_block(block)
        self.assertEqual(len(cc.atoms), 4)
        self.assertEqual(len(cc.rt.bonds), 3)
        self.assertAlmostEqual(cc.rt.get_bond("S", "O2").value, 1.496)
        self.assertAlmostEqual(cc.rt.get_bond("O3", "S").esd, 0.019)
        with self.assertRaises(RuntimeError):
            cc.rt.get_bond("O2", "O3")

    def test_shortest_path(self):
        path = full_path('HEM.cif')
        block = gemmi.cif.read(path)[-1]
        cc = gemmi.make_chemcomp_from_block(block)
        A = gemmi.Restraints.AtomId
        result = cc.rt.find_shortest_path(A('FE'), A('CMD'), [])
        self.assertEqual([aid.atom for aid in result],
                         ['FE', 'ND', 'C1D', 'C2D', 'CMD'])
        result = cc.rt.find_shortest_path(A('CMD'), A('FE'), [])
        self.assertEqual([aid.atom for aid in result],
                         ['CMD', 'C2D', 'C1D', 'ND', 'FE'])
        result = cc.rt.find_shortest_path(A('FE'), A('CMD'), [A('C1D')])
        self.assertEqual([aid.atom for aid in result],
                         ['FE', 'ND', 'C4D', 'C3D', 'C2D', 'CMD'])

class TestMonLib(unittest.TestCase):
    def test_path(self):
        path = full_path('')
        monlib = gemmi.read_monomer_lib(path, [])
        self.assertEqual(monlib.path('ALA'), path + "a/ALA.cif")
        self.assertEqual(monlib.path('CON'), path + "c/CON_CON.cif")

    @unittest.skipIf(os.getenv('CLIBD_MON') is None, "$CLIBD_MON not defined.")
    def test_read_monomer_lib(self):
        st = gemmi.read_structure(full_path('4oz7.pdb'))
        resnames = st[0].get_all_residue_names()
        monlib = gemmi.MonLib()
        ok = monlib.read_monomer_lib(os.environ['CLIBD_MON'], resnames)
        self.assertTrue(ok)
        monlib.update_old_atom_names(st)
        topo = gemmi.prepare_topology(st, monlib, model_index=0)
        topo = gemmi.prepare_topology(st, monlib, model_index=0,
                                      h_change=gemmi.HydrogenChange.Shift)
        self.assertIsNotNone(topo)
        # topo = gemmi.prepare_topology(st, monlib, model_index=0,
        #                              h_change=gemmi.HydrogenChange.ReAdd)
        # RuntimeError: Placing of hydrogen bonded to A/22W 6/N failed:
        # Missing angle restraint HN-N-C.


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import math
import pickle
import random
import unittest
import gemmi
try:
    from cctbx import sgtbx  # pytype: disable=import-error
    print('(w/ sgtbx)')
except ImportError:
    sgtbx = None

D = gemmi.Op.DEN

CANONICAL_SINGLES = {
    "x"     : [ D, 0, 0, 0],
    "z"     : [ 0, 0, D, 0],
    "-y"    : [ 0,-D, 0, 0],
    "-z"    : [ 0, 0,-D, 0],
    "x-y"   : [ D,-D, 0, 0],
    "-x+y"  : [-D, D, 0, 0],
    "x+1/2" : [ D, 0, 0, D//2],
    "y+1/4" : [ 0, D, 0, D//4],
    "z+3/4" : [ 0, 0, D, D*3//4],
    "z+1/3" : [ 0, 0, D, D*1//3],
    "z+1/6" : [ 0, 0, D, D//6],
    "z+2/3" : [ 0, 0, D, D*2//3],
    "z+5/6" : [ 0, 0, D, D*5//6],
    "-x+1/4": [-D, 0, 0, D//4],
    "-y+1/2": [ 0,-D, 0, D//2],
    "-y+3/4": [ 0,-D, 0, D*3//4],
    "-z+1/3": [ 0, 0,-D, D//3],
    "-z+1/6": [ 0, 0,-D, D//6],
    "-z+2/3": [ 0, 0,-D, D*2//3],
    "-z+5/6": [ 0, 0,-D, D*5//6],
}

OTHER_SINGLES = {
    # order and letter case may vary
    "Y-x"   : [-D, D, 0,  0],
    "-X"    : [-D, 0, 0,  0],
    "-1/2+Y": [ 0, D, 0, -D//2],
    # we want to handle non-crystallographic translations
    "x+3"   : [ D, 0, 0,  D*3],
    "1+Y"   : [ 0, D, 0,  D],
    "-2+Y"  : [ 0, D, 0, -D*2],
    "-z-5/6": [ 0, 0,-D, -D*5//6],
    # some CIF files from COD have decimal fractions
    "0.5-x":  [-D, 0, 0,  D//2],
    ".5+Y":   [ 0, D, 0, D//2],
    "-Z+0.25": [ 0, 0, -D, D//4],
    "x+0.6666666667": [ D, 0, 0, D*2//3],
    "1.25000-y": [ 0, -D, 0, D*5//4],
}

# From PLATON manual (WinGX v1.80, Chapter 9.3 PLATON),
# Table 7.1 Space group names known to the program
PLATON_SPACE_GROUP_NAMES = """\
P1 P-1 P2 P112 P21 P1121 C2 A2
B112 PM P11M PC PA PN P11B CM
AM B11M CC IC IA AA B11B P2/M
P112/M P21/M P1121/M C2/M A2/M B112/M P2/C P2/A
P112/B P21/C P21/A P2/N P1121/B P21/N P21/N11 C2/C
A2/A C2/N I2/C I2/N I2/M I2/A B112/B P222
P2221 P2212 P2122 P21212 P212121 C2221 B2212 C222
F222 I222 I212121 PMM2 PMC21 PCC2 PMA2 PCA21
PNC2 PMN21 PBA2 PNA21 P21NB PC21N PN21A PBN21
P21CN PNN2 CMM2 CMC21 CCC2 AMM2 ABM2 AMA2
ABA2 FMM2 FDD2 IMM2 IBA2 IMA2 PMMM PNNN
PCCM PBAN PMMA PNNA PMNA PCCA PBAM PCCN
PBCM PNNM PMMN PBCN PBCA PCAB PNMA PBNM
PMCN PNAM PMNB PCMN CMCM CMCA CMMM CCCM
CMMA CCCA FMMM FDDD IMMM IBAM IBCA IMMA
P4 P41 P42 P43 I4 I41 P-4 I-4
P4/M P42/M P4/N P42/N I4/M I41/A P422 P4212
P4122 P41212 P4222 P42212 P4322 P43212 I422 I4122
P4MM P4BM P42CM P42NM P4CC P4NC P42MC P42BC
I4MM I4CM I41MD I41CD P-42M P-42C P-421M P-421C
P-4M2 P-4C2 P-4B2 P-4N2 I-4M2 I-4C2 I-42M I-42D
P4/MMM P4/MCC P4/NBM P4/NNC P4/MBM P4/MNC P4/NMM P4/NCC
P42/MMC P42/MCM P42/NBC P42/NNM P42/MBC P42/MNM P42/NMC P42/NCM
I4/MMM I4/MCM I41/AMD I41/ACD P3 P31 P32 R3
R3R P-3 R-3 R-3R P312 P321 P3112 P3121
P3212 P3221 R32 R32R P3M1 P31M P3C1 P31C
R3M R3MR R3C R3CR P-31M P-31C P-3M1 P-3C1
R-3M R-3MR R-3C R-3CR P6 P61 P65 P62
P64 P63 P-6 P6/M P63/M P622 P6122 P6522
P6222 P6422 P6322 P6MM P6CC P63CM P63MC P-6M2
P-6C2 P-62M P-62C P6/MMM P6/MCC P63/MCM P63/MMC P23
F23 I23 P213 I213 PM3 PM-3 PN3 PN-3
FM3 FM-3 FD3 FD-3 IM3 IM-3 PA3 PA-3
IA3 IA-3 P432 P4232 F432 F4132 I432 P4332
P4132 I4132 P-43M F-43M I43M P-43N F-43C I-43D
PM3M PM-3M PN3N PN-3N PM3N PM-3N PN3M PN-3M
FM3M FM-3M FM3C FM-3C FD3M FD-3M FD3C FD-3C
IM3M IM-3M IA3D IA-3D
"""

def simple_monoclinic_unique_axis(hm):
    hm_split = hm.split()
    if len(hm_split) == 4 and hm_split.count('1') == 2:
        for i in [1,2,3]:
            if hm_split[i] != '1':
                return chr(ord('a') + i - 1)
    return '\0'

class TestSymmetry(unittest.TestCase):
    def test_parse_triplet_part(self):
        for single, row in CANONICAL_SINGLES.items():
            calculated = gemmi.parse_triplet_part(single)
            self.assertEqual(calculated, row)
        for single, row in OTHER_SINGLES.items():
            calculated = gemmi.parse_triplet_part(single)
            self.assertEqual(calculated, row)

    def test_triplet_roundtrip(self):
        singles = list(CANONICAL_SINGLES.keys())
        for i in range(4):
            items = [random.choice(singles) for j in range(3)]
            triplet = ','.join(items)
            op = gemmi.parse_triplet(triplet)
            self.assertEqual(op.triplet(), triplet)
            op2 = gemmi.seitz_to_op(op.float_seitz())
            self.assertEqual(op2.triplet(), triplet)
            op3 = gemmi.seitz_to_op(op.seitz())
            self.assertEqual(op3.triplet(), triplet)
        self.assertEqual(gemmi.Op(' x , - y, + z ').triplet(), 'x,-y,z')

    def test_triplet_style(self):
        op = gemmi.parse_triplet('A,-B , C')
        self.assertEqual(op.triplet(), 'a,-b,c')
        self.assertEqual(op.triplet('x'), 'x,-y,z')
        self.assertEqual(op.triplet('a'), 'a,-b,c')
        self.assertEqual(op.triplet('X'), 'X,-Y,Z')
        self.assertEqual(op.triplet('A'), 'A,-B,C')
        op = gemmi.parse_triplet('H,-k , L')
        self.assertEqual(op.triplet(), 'h,-k,l')
        self.assertEqual(op.triplet('h'), 'h,-k,l')
        self.assertEqual(op.triplet('H'), 'H,-K,L')

    def test_combine(self):
        a = gemmi.Op('x+1/3,z,-y')
        self.assertEqual(a.combine(a).triplet(), 'x+2/3,-y,-z')
        self.assertEqual('x,-y,z' * gemmi.Op('-x,-y,z'), '-x,y,z')
        a = gemmi.Op('-y+1/4,x+3/4,z+1/4')
        b = gemmi.Op('-x+1/2,y,-z')
        self.assertEqual((a * b).triplet(), '-y+1/4,-x+1/4,-z+1/4')
        c = '-y,-z,-x'
        self.assertNotEqual(b * c, c * b)
        self.assertEqual((a * c).triplet(), 'z+1/4,-y+3/4,-x+1/4')
        self.assertEqual(b * c, gemmi.Op('y+1/2,-z,x'))
        self.assertEqual(c * b, '-y,z,x+1/2')

    def test_invert(self):
        for xyz in ['-y,-x,-z+1/4', 'y,-x,z+3/4', 'y,x,-z', 'y+1/2,x,-z+1/3']:
            op = gemmi.Op(xyz)
            self.assertEqual(op * op.inverse(), 'x,y,z')
            self.assertEqual(op.inverse().inverse(), op)
        # test change-of-basis op between hexagonal and trigonal settings
        op = gemmi.Op("-y+z,x+z,-x+y+z")  # det=3
        self.assertEqual(op.det_rot(), 3 * gemmi.Op.DEN**3)
        inv = op.inverse()
        self.assertEqual(inv * op, 'x,y,z')
        self.assertEqual(op * inv, 'x,y,z')
        expected_inv = '-x/3+2/3*y-z/3,-2/3*x+y/3+z/3,x/3+y/3+z/3'
        self.assertEqual(inv.triplet(), expected_inv)
        self.assertEqual(gemmi.Op(expected_inv), inv)
        op = gemmi.Op('1/2*x+1/2*y,-1/2*x+1/2*y,z')
        self.assertEqual(op.inverse().triplet(), 'x-y,x+y,z')
        # check also alternative writing
        op2 = gemmi.parse_triplet('a/2+b/2,-A/2+b/2,C', notation='a')
        self.assertEqual(op, op2)

    def test_rot_type(self):
        # examples taken from tst_sgtbx.py in cctbx
        self.assertEqual(gemmi.Op('z,y,-x').rot_type(), 4)
        self.assertEqual(gemmi.Op('x,y,z').rot_type(), 1)
        self.assertEqual(gemmi.Op('2*x,y,z').rot_type(), 0)
        self.assertEqual(gemmi.Op('-x,z-y,-y').rot_type(), -6)
        self.assertEqual(gemmi.Op('-x,z-y,-y').inverse().rot_type(), -6)

    def test_generators_from_hall(self):
        # first test on example matrices from
        # http://cci.lbl.gov/sginfo/hall_symbols.html
        self.assertEqual(gemmi.generators_from_hall('p -2xc').sym_ops,
                         ['x,y,z', '-x,y,z+1/2'])
        self.assertEqual(gemmi.generators_from_hall('p 3*').sym_ops,
                         ['x,y,z', 'z,x,y'])
        self.assertEqual(gemmi.generators_from_hall('p 4vw').sym_ops,
                         ['x,y,z', '-y,x+1/4,z+1/4'])
        self.assertEqual(gemmi.generators_from_hall('p 61 2 (0 0 -1)').sym_ops,
                         ['x,y,z', 'x-y,x,z+1/6', '-y,-x,-z+5/6'])
        # then on examples from the 530 settings
        self.assertEqual(gemmi.generators_from_hall('P -2 -2').sym_ops,
                         ['x,y,z', 'x,y,-z', '-x,y,z'])
        # the same operations in different notation
        a = gemmi.generators_from_hall('P 3*')
        b = gemmi.generators_from_hall('R 3 (-y+z,x+z,-x+y+z)')
        self.assertEqual(a.sym_ops, b.sym_ops)
        self.assertEqual(a.cen_ops, b.cen_ops)

    def compare_hall_symops_with_sgtbx(self, hall, existing_group=True):
        cctbx_sg = sgtbx.space_group(hall)
        cctbx_triplets = set(m.as_xyz() for m in cctbx_sg.all_ops(mod=1))
        gemmi_gops = gemmi.symops_from_hall(hall)
        self.assertEqual(len(gemmi_gops.sym_ops), cctbx_sg.order_p())
        self.assertEqual(len(gemmi_gops.cen_ops), cctbx_sg.n_ltr())
        self.assertEqual(len(gemmi_gops), cctbx_sg.order_z())
        self.assertEqual(gemmi_gops.is_centrosymmetric(), cctbx_sg.is_centric())
        ctr = gemmi_gops.find_centering()
        self.assertEqual(ctr, cctbx_sg.conventional_centring_type_symbol())
        gemmi_triplets = set(m.triplet() for m in gemmi_gops)
        self.assertEqual(cctbx_triplets, gemmi_triplets)
        gemmi_sg = gemmi.find_spacegroup_by_ops(gemmi_gops)
        if existing_group:
            self.assertEqual(gemmi_sg.point_group_hm(),
                             cctbx_sg.point_group_type())
            self.assertEqual(gemmi_sg.crystal_system_str(),
                             cctbx_sg.crystal_system().lower())
            self.assertEqual(gemmi_sg.is_sohncke(), cctbx_sg.is_chiral())
            self.assertEqual(gemmi_sg.is_enantiomorphic(),
                             cctbx_sg.type().is_enantiomorphic())
            self.assertEqual(gemmi_sg.is_symmorphic(),
                             cctbx_sg.type().is_symmorphic())
            self.assertEqual(gemmi_sg.centring_type(), ctr)
        else:
            self.assertIsNone(gemmi_sg)

    def test_with_sgtbx(self):
        if sgtbx is None:
            return
        for s in gemmi.spacegroup_table():
            self.compare_hall_symops_with_sgtbx(s.hall)
        self.compare_hall_symops_with_sgtbx('C -4 -2b', existing_group=False)

    def test_table(self):
        for sg in gemmi.spacegroup_table():
            if sg.ccp4 != 0:
                self.assertEqual(sg.ccp4 % 1000, sg.number)
            is_laue = (sg.laue_str() == sg.point_group_hm())
            self.assertEqual(sg.operations().is_centrosymmetric(), is_laue)
            self.assertEqual(sg.is_centrosymmetric(), is_laue)
            if sgtbx:
                hall = sg.hall.encode()
                cctbx_sg = sgtbx.space_group(hall)
                cctbx_info = sgtbx.space_group_info(group=cctbx_sg)
                self.assertEqual(sg.is_reference_setting(),
                                 cctbx_info.is_reference_setting())
                #to_ref = cctbx_info.change_of_basis_op_to_reference_setting()
                #from_ref = '%s' % cob_to_ref.inverse().c()
                c2p_sg = gemmi.Op(cctbx_sg.z2p_op().c().inverse().as_xyz())
                self.assertEqual(sg.centred_to_primitive(), c2p_sg)
                hand_sgtbx = cctbx_info.change_of_basis_op_to_other_hand()
                self.assertEqual(sg.change_of_hand_op().triplet(),
                                 hand_sgtbx.as_xyz())
            ops = gemmi.get_spacegroup_reference_setting(sg.number).operations()
            ops.change_basis_forward(sg.basisop)
            self.assertEqual(ops, sg.operations(), msg=sg.xhm())
            self.assertEqual(sg.monoclinic_unique_axis(),
                             simple_monoclinic_unique_axis(sg.hm))
        itb = gemmi.spacegroup_table_itb()
        if sgtbx:
            for s in sgtbx.space_group_symbol_iterator():
                self.assertEqual(s.hall().strip(), next(itb).hall)
            with self.assertRaises(StopIteration):
                next(itb)

    def test_enantiomorphic_pairs(self):
        counter = 0
        for sg in gemmi.spacegroup_table():
            coh = sg.change_of_hand_op()
            ops = sg.operations()
            ops.change_basis_forward(coh)
            other_hand = gemmi.find_spacegroup_by_ops(ops)
            self.assertIsNotNone(other_hand, msg=f'{sg} {coh}')
            if sg.hall != other_hand.hall:
                counter += 1
            elif sg != other_hand:
                # duplicates
                self.assertTrue(sg.number == 68 or sg.xhm() == 'A b a m')
        self.assertEqual(counter, 2 * 11)

    def test_symmorphic(self):
        for sg in gemmi.spacegroup_table():
            ops = sg.operations()
            symmor_ops = ops.derive_symmorphic()
            self.assertEqual(len(ops), len(symmor_ops))
            symmor = gemmi.find_spacegroup_by_ops(symmor_ops)
            self.assertTrue(symmor is not None, msg=sg.xhm())
            self.assertTrue(symmor.is_symmorphic())
            if sg.is_symmorphic():
                self.assertEqual(sg.number, symmor.number)

    def test_find_spacegroup(self):
        self.assertEqual(gemmi.SpaceGroup('P21212').hm, 'P 21 21 2')
        self.assertEqual(gemmi.find_spacegroup_by_name('P21').hm, 'P 1 21 1')
        self.assertEqual(gemmi.find_spacegroup_by_name('P 2').hm, 'P 1 2 1')
        def check_xhm(name, xhm):
            self.assertEqual(gemmi.SpaceGroup(name).xhm(), xhm)
        check_xhm('R 3 2', 'R 3 2:H')
        check_xhm('R 3 2:h', 'R 3 2:H')
        check_xhm('R32:H', 'R 3 2:H')
        check_xhm('H32', 'R 3 2:H')
        check_xhm(' R32:H ', 'R 3 2:H')
        check_xhm('     H32', 'R 3 2:H')
        check_xhm('R 3 2:R', 'R 3 2:R')
        check_xhm('P6', 'P 6')
        check_xhm('P 6', 'P 6')
        check_xhm('P65', 'P 65')
        check_xhm('I1211', 'I 1 21 1')
        check_xhm('Aem2', 'A b m 2')
        check_xhm('C c c e', 'C c c a:1')
        check_xhm('i2', 'I 1 2 1')
        check_xhm('I 41/A', 'I 41/a:1')
        check_xhm('I -4 2 D', 'I -4 2 d')
        check_xhm('P 1 21/c 1', 'P 1 21/c 1')
        check_xhm('P 21 21 2 A', 'P 21212(a)')
        check_xhm('B 2', 'B 1 1 2')
        self.assertRaises(ValueError, gemmi.SpaceGroup, 'i3')
        self.assertEqual(gemmi.find_spacegroup_by_number(5).hm, 'C 1 2 1')
        self.assertEqual(gemmi.SpaceGroup(4005).hm, 'I 1 2 1')
        self.assertIsNone(gemmi.find_spacegroup_by_name('abc'))

    def test_identity(self):
        # Checks that all functions return a reference, not a copy
        p1 = gemmi.find_spacegroup_by_number(1)
        self.assertTrue(gemmi.find_spacegroup_by_name('P1') is p1)
        #self.assertTrue(gemmi.SpaceGroup(1) is p1)
        #self.assertTrue(gemmi.SpaceGroup('P 1') is p1)
        gops = gemmi.GroupOps([gemmi.Op()])
        self.assertTrue(gemmi.find_spacegroup_by_ops(gops) is p1)
        self.assertTrue(next(gemmi.spacegroup_table()) is p1)
        self.assertTrue(next(gemmi.spacegroup_table_itb()) is p1)

        a = gemmi.SpaceGroup("C c c a:1")
        b = gemmi.SpaceGroup("C c c b:1")
        self.assertTrue(a == b)
        self.assertTrue(a is not b)

    def test_groupops(self):
        gops = gemmi.GroupOps([gemmi.Op(t) for t in ['x, y, z',
                                                     'x, -y, z+1/2',
                                                     'x+1/2, y+1/2, z',
                                                     'x+1/2, -y+1/2, z+1/2']])
        self.assertEqual(gops.find_centering(), 'C')
        self.assertEqual(len(gops), 4)
        self.assertEqual(gemmi.find_spacegroup_by_ops(gops).hm, 'C 1 c 1')

    def test_add_inversion(self):
        gops = gemmi.SpaceGroup(1).operations()
        self.assertTrue(gops.add_inversion())
        self.assertEqual(len(gops), 2)
        self.assertEqual(gops.sym_ops, [gemmi.Op(), gemmi.Op('-x,-y,-z')])
        self.assertFalse(gops.add_inversion())

    def change_basis(self, name_a, name_b, basisop_triplet):
        basisop = gemmi.Op(basisop_triplet)
        a = gemmi.find_spacegroup_by_name(name_a)
        b = gemmi.find_spacegroup_by_name(name_b)
        ops = a.operations()
        ops.change_basis_forward(basisop)
        self.assertEqual(ops, b.operations())
        ops.change_basis_backward(basisop)
        self.assertEqual(ops, a.operations())

    def test_change_basis(self):
        self.change_basis('I2', 'C2', 'x,y,x+z')
        self.change_basis('C 1 c 1', 'C 1 n 1', 'x+1/4,y+1/4,z')
        self.change_basis('R 3 :H', 'R 3 :R', '-y+z,x+z,-x+y+z')
        self.change_basis('A -1', 'P -1', '-x,-y+z,y+z')

    def test_short_name(self):
        for (longer, shorter) in [('P 21 2 21', 'P21221'),
                                  ('P 1 2 1',   'P2'),
                                  ('P 1',       'P1'),
                                  ('R 3 2:R',   'R32'),
                                  ('R 3 2:H',   'H32')]:
            self.assertEqual(gemmi.SpaceGroup(longer).short_name(), shorter)

    def compare_short_names_with_symop_lib():
        for line in open('symop.lib'):
            if line and not line[0].isspace():
                fields = line.partition('!')[0].split(None, 6)
                #spacegroups = shlex.split(fields[-1])
                g = gemmi.find_spacegroup_by_number(int(fields[0]))
                if fields[3] != g.short_name():
                    print('[%s] %s %s' % (g.xhm(), g.short_name(), fields[3]))

    def test_operations(self):
        gops = gemmi.symops_from_hall('-P 2a 2ac (z,x,y)')
        self.assertEqual(set(gemmi.SpaceGroup('Pbaa').operations()), set(gops))
        self.assertEqual(gemmi.find_spacegroup_by_ops(gops).hm, 'P b a a')

    def test_find_grid_factors(self):
        def fact(name):
            return gemmi.SpaceGroup(name).operations().find_grid_factors()
        self.assertEqual(fact('P21'), [1, 2, 1])
        self.assertEqual(fact('P61'), [1, 1, 6])

    # based on example from pages 9-10 in
    # https://www.iucr.org/education/pamphlets/9
    def test_phase_shift(self):
        ops = gemmi.find_spacegroup_by_name('P 31 2 1').operations()
        refl = [3, 0, 1]
        expected_equiv = [
            # in the paper the last two reflections are swapped
            [3, 0, 1], [0, -3, 1], [-3, 3, 1],
            [0, 3, -1], [3, -3, -1], [-3, 0, -1]]
        self.assertEqual([op.apply_to_hkl(refl) for op in ops], expected_equiv)
        expected_shifts = [0, -120, -240, 0, -240, -120]
        for op, expected in zip(ops, expected_shifts):
            shift = math.degrees(op.phase_shift(refl))
            self.assertAlmostEqual((shift - expected) % 360, 0)

    def test_reciprocal_asu_checker(self):
        sg = gemmi.SpaceGroup('I 1 2 1')
        checker = gemmi.ReciprocalAsu(sg)
        self.assertTrue(checker.is_in([-5, 5, 1]))
        self.assertFalse(checker.is_in([5, 5, -1]))
        checker = gemmi.ReciprocalAsu(sg, tnt=True)
        self.assertTrue(checker.is_in([5, 5, -1]))
        self.assertFalse(checker.is_in([-5, 5, 1]))

    def test_reflection_properties(self):
        sg = gemmi.SpaceGroup('I 1 2 1')
        gops = sg.operations()
        self.assertTrue(gops.is_reflection_centric([3,0,3]))
        self.assertFalse(gops.is_reflection_centric([3,3,3]))
        self.assertEqual(gops.epsilon_factor([3,0,3]), 2)
        self.assertEqual(gops.epsilon_factor([0,3,0]), 4)
        self.assertFalse(gops.is_systematically_absent([1,2,3]))
        self.assertTrue(gops.is_systematically_absent([1,2,4]))
        sg = gemmi.SpaceGroup('F 4 3 2')
        gops = sg.operations()
        self.assertEqual(gops.epsilon_factor([2,0,0]), 16)
        self.assertEqual(gops.epsilon_factor([3,3,3]), 12)
        self.assertEqual(gops.epsilon_factor_without_centering([2,0,0]), 4)

    def test_pickling(self):
        sg = gemmi.SpaceGroup("P 31 2 1")
        pkl_string = pickle.dumps(sg, protocol=pickle.HIGHEST_PROTOCOL)
        result = pickle.loads(pkl_string)
        self.assertTrue(isinstance(result, gemmi.SpaceGroup))
        self.assertEqual(sg.xhm(), result.xhm())

    # test space group names listed in PLATON docs
    def test_platon_spacegroup_names(self):
        if sgtbx is None:
            return
        for s in PLATON_SPACE_GROUP_NAMES.split():
            # these names are not recognized neither by sgtbx nor by gemmi
            if s in ['I2/N', 'I43M']:
                continue
            # By default, for groups such as Pnnn, gemmi uses the first
            # settings (ext==1) and sgtbx the ones with ext==2.
            # But with table_id='A', sgtbx also uses extension 1.
            sgtbx_group = sgtbx.space_group_type(s, table_id='A')
            sgtbx_ops = gemmi.symops_from_hall(sgtbx_group.hall_symbol())
            gemmi_ops = gemmi.SpaceGroup(s).operations()
            self.assertEqual(sgtbx_ops, gemmi_ops)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

import unittest
import gemmi
from gemmi import Element, IT92_get_exact, IT92_set_ignore_charge


class TestElem(unittest.TestCase):
    def test_name(self):
        for name in ELEMENT_MASS:
            el = Element(name)
            self.assertEqual(el.name, name)
        self.assertEqual(Element('AL').name, 'Al')
        self.assertEqual(Element('al').name, 'Al')
        # We check only the first two characters now.
        self.assertEqual(Element('alt').name, 'Al')
        self.assertEqual(Element('Q').name, 'X')
        self.assertEqual(Element('QQ').name, 'X')
        self.assertEqual(Element('--').name, 'X')
        self.assertEqual(Element('').name, 'X')

    def test_weight(self):
        for name, mass in ELEMENT_MASS.items():
            self.assertAlmostEqual(Element(name).weight, mass, delta=1e-3)

    def test_atomic_number(self):
        self.assertEqual(Element('O').atomic_number, 8)
        self.assertEqual(Element('D').atomic_number, 1)
        self.assertEqual(Element('X').atomic_number, 0)
        for n in range(119):
            self.assertEqual(Element(n).atomic_number, n)

    def test_properties(self):
        self.assertEqual(Element('K').vdw_r, 2.75)

    def test_coef(self):
        h = Element('H')
        d = Element('D')
        fe = Element('Fe')
        cf = Element('Cf')
        es = Element('Es')
        self.assertIsNotNone(cf.it92)
        self.assertIsNotNone(cf.c4322)
        self.assertIsNone(es.it92)
        self.assertIsNone(es.c4322)
        self.assertEqual(h.it92.get_coefs(), d.it92.get_coefs())
        self.assertEqual(h.c4322.get_coefs(), d.c4322.get_coefs())
        self.assertEqual(Element('X').it92.get_coefs(),
                         Element('O').it92.get_coefs())
        fe_coefs = fe.it92.get_coefs()
        fe0_coefs = IT92_get_exact(fe, 0).get_coefs()
        IT92_set_ignore_charge(False)
        fe2_coefs = IT92_get_exact(fe, 2).get_coefs()
        self.assertIsNone(IT92_get_exact(fe, 4))
        self.assertEqual(fe_coefs, fe0_coefs)
        self.assertNotEqual(fe_coefs, fe2_coefs)

    def test_resinfo(self):
        for i in gemmi.resinfo_table():
            ri = gemmi.find_tabulated_residue(i.name)
            self.assertEqual(ri.name, i.name)


ELEMENT_MASS = {
    'H': 1.00794, 'He': 4.0026, 'Li': 6.941, 'Be': 9.012182, 'B': 10.811,
    'C': 12.0107, 'N': 14.0067, 'O': 15.9994, 'F': 18.998403, 'Ne': 20.1797,
    'Na': 22.98977, 'Mg': 24.305, 'Al': 26.981539, 'Si': 28.0855,
    'P': 30.973761, 'S': 32.065, 'Cl': 35.453, 'Ar': 39.948,
    'K': 39.0983, 'Ca': 40.078, 'Sc': 44.95591, 'Ti': 47.867, 'V': 50.9415,
    'Cr': 51.9961, 'Mn': 54.93805, 'Fe': 55.845, 'Co': 58.9332,
    'Ni': 58.6934, 'Cu': 63.546, 'Zn': 65.38, 'Ga': 69.723, 'Ge': 72.64,
    'As': 74.9216, 'Se': 78.96, 'Br': 79.904, 'Kr': 83.798,
    'Rb': 85.4678, 'Sr': 87.62, 'Y': 88.90585, 'Zr': 91.224, 'Nb': 92.9064,
    'Mo': 95.95, 'Tc': 98, 'Ru': 101.07, 'Rh': 102.9055, 'Pd': 106.42,
    'Ag': 107.8682, 'Cd': 112.411, 'In': 114.818, 'Sn': 118.71, 'Sb': 121.76,
    'Te': 127.6, 'I': 126.90447, 'Xe': 131.293, 'Cs': 132.905, 'Ba': 137.327,
    'La': 138.905, 'Ce': 140.116, 'Pr': 140.908, 'Nd': 144.24, 'Pm': 145,
    'Sm': 150.36, 'Eu': 151.964, 'Gd': 157.25, 'Tb': 158.925, 'Dy': 162.5,
    'Ho': 164.93, 'Er': 167.259, 'Tm': 168.934, 'Yb': 173.05, 'Lu': 174.967,
    'Hf': 178.49, 'Ta': 180.948, 'W': 183.84, 'Re': 186.207, 'Os': 190.23,
    'Ir': 192.217, 'Pt': 195.084, 'Au': 196.967, 'Hg': 200.59, 'Tl': 204.383,
    'Pb': 207.2, 'Bi': 208.98, 'Po': 209, 'At': 210, 'Rn': 222,
    'Fr': 223, 'Ra': 226, 'Ac': 227, 'Th': 232.038, 'Pa': 231.036,
    'U': 238.029, 'Np': 237, 'Pu': 244, 'Am': 243, 'Cm': 247,
    'Bk': 247, 'Cf': 251, 'Es': 252, 'Fm': 257, 'Md': 258,
    'No': 259, 'Lr': 262, 'Rf': 267, 'Db': 268, 'Sg': 271,
    'Bh': 272, 'Hs': 270, 'Mt': 276,
    'D': 2.0141, 'X': 1,
}

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3

# Compares graphs of molecules from cif files (Refmac dictionary or similar)
# with CCD entries.

import sys
import networkx
from networkx.algorithms import isomorphism
import gemmi

CCD_PATH = 'components.cif.gz'

def graph_from_chemcomp(cc):
    G = networkx.Graph()
    for atom in cc.atoms:
        G.add_node(atom.id, Z=atom.el.atomic_number)
    for bond in cc.rt.bonds:
        G.add_edge(bond.id1.atom, bond.id2.atom)
    return G

def compare(cc1, cc2):
    s1 = {a.id for a in cc1.atoms}
    s2 = {a.id for a in cc2.atoms}
    b1 = {b.lexicographic_str() for b in cc1.rt.bonds}
    b2 = {b.lexicographic_str() for b in cc2.rt.bonds}
    if s1 == s2 and b1 == b2:
        #print(cc1.name, "the same")
        return
    G1 = graph_from_chemcomp(cc1)
    G2 = graph_from_chemcomp(cc2)
    node_match = isomorphism.categorical_node_match('Z', 0)
    GM = isomorphism.GraphMatcher(G1, G2, node_match=node_match)
    if GM.is_isomorphic():
        print(cc1.name, 'is isomorphic')
        # we could use GM.match(), but here we try to find the shortest diff
        short_diff = None
        for n, mapping in enumerate(GM.isomorphisms_iter()):
            diff = {k: v for k, v in mapping.items() if k != v}
            if short_diff is None or len(diff) < len(short_diff):
                short_diff = diff
            if n == 10000:  # don't spend too much here
                print(' (it may not be the simplest isomorphism)')
                break
        assert short_diff is not None
        for id1, id2 in short_diff.items():
            print('\t', id1, '->', id2)
    else:
        print(cc1.name, 'differs')
        if s2 - s1:
            print('\tmissing:', ' '.join(s2 - s1))
        if s1 - s2:
            print('\textra:  ', ' '.join(s1 - s2))

def main():
    ccd = gemmi.cif.read(CCD_PATH)
    absent = 0
    for f in sys.argv[1:]:
        block = gemmi.cif.read(f)[-1]
        cc1 = gemmi.make_chemcomp_from_block(block)
        try:
            block2 = ccd[cc1.name]
        except KeyError:
            absent += 1
            #print(cc1.name, 'not in CCD')
            continue
        cc2 = gemmi.make_chemcomp_from_block(block2)
        cc1.remove_hydrogens()
        cc2.remove_hydrogens()
        compare(cc1, cc2)
    if absent != 0:
        print(absent, 'of', len(sys.argv) - 1, 'monomers not found in CCD')

main()



#!/usr/bin/env python

import os
import unittest
import gemmi
from common import full_path, get_path_for_tempfile, assert_numpy_equal, numpy

def compare_maps(self, a, b, atol):
    #print(abs(numpy.array(a) - b).max())
    self.assertTrue(numpy.allclose(a, b, atol=atol, rtol=0))

def compare_asu_data(self, asu_data, data, f, phi):
    asu_dict = {tuple(a.hkl): a.value for a in asu_data}
    data_hkl = data.make_miller_array()
    self.assertEqual(data_hkl.dtype, 'int32')
    if type(data) is gemmi.ReflnBlock:
        data_f = data.make_float_array(f)
        data_phi = data.make_float_array(phi)
    else:
        data_f = data.column_with_label(f).array
        data_phi = data.column_with_label(phi).array
    self.assertTrue(len(data_f) > 100)
    asu_val = numpy.array([asu_dict[tuple(hkl)] for hkl in data_hkl])
    self.assertTrue(numpy.allclose(data_f, abs(asu_val), atol=1e-4, rtol=0))
    x180 = abs(180 - abs(data_phi - numpy.angle(asu_val, deg=True)))
    self.assertTrue(numpy.allclose(x180, 180, atol=6e-2, rtol=0.))

def fft_test(self, data, f, phi, size, order=gemmi.AxisOrder.XYZ):
    if numpy is None:
        return
    self.assertTrue(data.data_fits_into(size))
    grid_full = data.get_f_phi_on_grid(f, phi, size, half_l=False, order=order)
    self.assertEqual(grid_full.axis_order, order)
    array_full = grid_full.array
    map1 = gemmi.transform_f_phi_grid_to_map(grid_full)
    self.assertEqual(map1.axis_order, order)
    map2 = numpy.fft.ifftn(array_full.conj())
    map2 = numpy.real(map2) * (map2.size / grid_full.unit_cell.volume)
    compare_maps(self, map1, map2, atol=1e-6)
    map3 = data.transform_f_phi_to_map(f, phi, size, order=order)
    compare_maps(self, map1, map3, atol=6e-7)

    grid2 = gemmi.transform_map_to_f_phi(map1, half_l=False)
    self.assertFalse(grid2.half_l)
    self.assertEqual(grid2.axis_order, order)
    compare_maps(self, grid2, array_full, atol=2e-4)
    if grid2.axis_order != gemmi.AxisOrder.ZYX:
        compare_asu_data(self, grid2.prepare_asu_data(), data, f, phi)
    grid_half = data.get_f_phi_on_grid(f, phi, size, half_l=True, order=order)
    if order == gemmi.AxisOrder.ZYX:  # half_l+ZYX not supported yet
        return
    grid3 = gemmi.transform_map_to_f_phi(map1, half_l=True)
    self.assertTrue(grid3.half_l)
    self.assertEqual(grid3.axis_order, order)
    compare_maps(self, grid3, grid_half, atol=2e-4)
    compare_asu_data(self, grid3.prepare_asu_data(), data, f, phi)

    asu_data = grid_full.prepare_asu_data()
    back_grid = asu_data.get_f_phi_on_grid(size, half_l=False, order=order)
    compare_maps(self, back_grid, grid_full, atol=4e-5)

    asu_data = grid3.prepare_asu_data()
    back_grid = asu_data.get_f_phi_on_grid(size, half_l=True, order=order)
    compare_maps(self, back_grid, grid3, atol=5e-5)

    asu_data = data.get_f_phi(f, phi)
    compare_asu_data(self, asu_data, data, f, phi)
    hkl_grid = asu_data.get_f_phi_on_grid(size, order=order)
    compare_maps(self, hkl_grid, grid_full, atol=1e-4)
    if order != gemmi.AxisOrder.ZYX:
        m1 = gemmi.transform_f_phi_grid_to_map(hkl_grid)
        m2 = asu_data.transform_f_phi_to_map(exact_size=size)
        compare_maps(self, m1, m2, atol=1e-6)


class TestMtz(unittest.TestCase):
    def test_read_write(self):
        path = full_path('5e5z.mtz')
        mtz = gemmi.read_mtz_file(path)
        self.assertEqual(mtz.spacegroup.hm, 'P 1 21 1')
        out_name = get_path_for_tempfile()
        mtz.write_to_file(out_name)
        mtz_bytes = mtz.write_to_bytes()
        with open(out_name, 'rb') as f:
            self.assertEqual(mtz_bytes, f.read())
        mtz2 = gemmi.read_mtz_file(out_name)
        mtz3 = gemmi.read_mtz_file(out_name, with_data=False)
        os.remove(out_name)
        self.assertEqual(mtz2.spacegroup.hm, 'P 1 21 1')
        self.assertEqual(mtz3.spacegroup.hm, 'P 1 21 1')
        if numpy is not None:
            assert_numpy_equal(self, mtz.array, mtz2.array)
            self.assertEqual(mtz3.array.shape, (0, 8))

    def test_remove_and_add_column(self):
        path = full_path('5e5z.mtz')
        col_name = 'FREE'
        mtz = gemmi.read_mtz_file(path)
        col = mtz.column_with_label(col_name)
        col_idx = col.idx
        ncol = len(mtz.columns)
        if numpy is None:
            return
        arr = col.array.copy()
        mtz_data = mtz.array.copy()
        self.assertEqual(mtz_data.shape, (mtz.nreflections, ncol))
        mtz.remove_column(col_idx)
        self.assertEqual(len(mtz.columns), ncol-1)
        self.assertEqual(mtz.array.shape,
                         (mtz.nreflections, ncol-1))
        col = mtz.add_column(col_name, 'I', dataset_id=0, pos=col_idx)
        col.array[:] = arr
        assert_numpy_equal(self, mtz_data, mtz.array)

    def asu_data_test(self, grid):
        asu = grid.prepare_asu_data()
        d = asu.make_d_array()
        asu2 = gemmi.ComplexAsuData(asu.unit_cell,
                                    asu.spacegroup,
                                    asu.miller_array[d > 2.5],
                                    asu.value_array[d > 2.5])
        ngrid = asu2.transform_f_phi_to_map(sample_rate=1.5)
        hkl_grid = asu2.get_f_phi_on_grid([ngrid.nu, ngrid.nv, ngrid.nw])
        alt_ngrid = gemmi.transform_f_phi_grid_to_map(hkl_grid)
        compare_maps(self, ngrid, alt_ngrid, atol=1e-6)

    def test_f_phi_grid(self):
        path = full_path('5wkd_phases.mtz.gz')
        mtz = gemmi.read_mtz_file(path)
        size = mtz.get_size_for_hkl()
        for half_l in (False, True):
            grid1 = mtz.get_f_phi_on_grid('FWT', 'PHWT', size, half_l=half_l)
            grid2 = mtz.get_f_phi_on_grid('FWT', 'PHWT', size, half_l=half_l,
                                          order=gemmi.AxisOrder.ZYX)
            if numpy is None:
                continue
            array1 = grid1.array
            array2 = grid2.array
            self.assertTrue((array2 == array1.transpose(2,1,0)).all())
            self.asu_data_test(grid1)

        fft_test(self, mtz, 'FWT', 'PHWT', size)
        fft_test(self, mtz, 'FWT', 'PHWT', size, order=gemmi.AxisOrder.ZYX)

    def test_value_grid(self):
        #path = full_path('5wkd_phases.mtz.gz')
        path = full_path('5e5z.mtz')
        mtz = gemmi.read_mtz_file(path)
        size = mtz.get_size_for_hkl()
        if numpy is None:
            return
        asu = gemmi.ReciprocalAsu(mtz.spacegroup)
        mtz_data = mtz.array
        fp_idx = mtz.column_labels().index('FP')
        fp_map = {}
        for row in mtz_data:
            fp_map[tuple(row[0:3])] = row[fp_idx]
        for order in (gemmi.AxisOrder.XYZ, gemmi.AxisOrder.ZYX):
            for half_l in (True, False):
                grid = mtz.get_value_on_grid('FP', size,
                                             half_l=half_l, order=order)
                counter = 0
                for point in grid:
                    hkl = grid.to_hkl(point)
                    value = fp_map.get(tuple(hkl))
                    if asu.is_in(hkl):
                        if value is not None:
                            self.assertTrue(point.value == value
                                            or (numpy.isnan(point.value)
                                                and numpy.isnan(value)))
                            counter += 1
                        else:
                            self.assertEqual(point.value, 0.)
                    else:
                        self.assertIsNone(value)
                self.assertEqual(counter, mtz_data.shape[0])

    def test_batch_headers(self):
        b = gemmi.Mtz.Batch()
        self.assertEqual(b.ints[12], 0)
        b.ints[12] = 555
        self.assertEqual(b.ints[12], 555)
        self.assertEqual(b.floats[3], 0)
        b.floats[3] = 1234.5
        self.assertEqual(b.floats[3], 1234.5)
        another_batch = gemmi.Mtz.Batch()
        another_batch.ints = b.ints
        self.assertEqual(another_batch.ints[12], 555)
        another_batch.floats = b.floats
        self.assertEqual(another_batch.floats[3], 1234.5)


class TestSfMmcif(unittest.TestCase):
    def test_reading(self):
        doc = gemmi.cif.read(full_path('r5wkdsf.ent'))
        rblock = gemmi.as_refln_blocks(doc)[0]
        self.assertEqual(rblock.spacegroup.hm, 'C 1 2 1')

        size = rblock.get_size_for_hkl()
        for order in (gemmi.AxisOrder.XYZ, gemmi.AxisOrder.ZYX):
            fft_test(self, rblock, 'pdbx_FWT', 'pdbx_PHWT', size)

    def test_scaling(self):
        doc = gemmi.cif.read(full_path('r5wkdsf.ent'))
        rblock = gemmi.as_refln_blocks(doc)[0]
        fobs_data = rblock.get_value_sigma('F_meas_au', 'F_meas_sigma_au')
        if numpy:
            self.assertEqual(fobs_data.value_array.shape, (367, 2))

        # without mask
        fc_data = rblock.get_f_phi('F_calc_au', 'phase_calc')
        scaling = gemmi.Scaling(fc_data.unit_cell, fc_data.spacegroup)
        scaling.prepare_points(fc_data, fobs_data)
        scaling.fit_isotropic_b_approximately()
        scaling.fit_parameters()
        #print(scaling.k_overall, scaling.b_overall)
        scaling.scale_data(fc_data)

        # with mask
        # TODO
        #fc_data = rblock.get_f_phi('F_calc_au', 'phase_calc')
        #st = gemmi.read_structure(full_path('5wkd.pdb'))
        #fmask = ...

class TestBinner(unittest.TestCase):
    def test_binner(self):
        path = full_path('5e5z.mtz')
        def check_limits_17(limits):
            self.assertEqual(len(limits), 17)
            self.assertAlmostEqual(limits[10], 0.27026277234462415)
        mtz = gemmi.read_mtz_file(path)
        binner = gemmi.Binner()
        method = gemmi.Binner.Method.Dstar3
        binner.setup(17, method, mtz)
        check_limits_17(binner.limits)
        self.assertEqual(binner.size, 17)
        binner = gemmi.Binner()
        binner.setup(17, method, mtz, cell=mtz.cell)
        check_limits_17(binner.limits)
        self.assertEqual(binner.get_bin([3,3,3]), 9)
        if numpy is None:
            return
        binner.setup(17, method, mtz.make_miller_array(), cell=mtz.cell)
        check_limits_17(binner.limits)
        binner.setup_from_1_d2(17, method, mtz.make_1_d2_array(), mtz.cell)
        check_limits_17(binner.limits)
        hkls = [[0,0,1], [3,3,3], [10,10,10]]
        bins = [0,9,16]
        # TODO: remove explicit call to numpy.array after this is implemented:
        # https://github.com/wjakob/nanobind/discussions/327
        self.assertEqual(list(binner.get_bins(numpy.array(hkls))), bins)
        # TODO: remove explicit call to numpy.array, as above
        inv_d2 = numpy.array([mtz.cell.calculate_1_d2(h) for h in hkls])
        self.assertEqual(list(binner.get_bins_from_1_d2(inv_d2)), bins)

class TestMerging(unittest.TestCase):
    def test_reading(self):
        doc = gemmi.cif.read(full_path('cc12-hkl.cif'))
        rblock = gemmi.as_refln_blocks(doc)[0]
        intens = gemmi.Intensities()
        intens.import_refln_block(rblock)
        intens.prepare_for_merging(gemmi.DataType.Anomalous)
        self.assertEqual(len(intens), 12)
        self.assertEqual(intens.spacegroup.xhm(), 'P 2 3')
        y_stats = intens.calculate_merging_stats(None, use_weights='Y')[0]
        u_stats = intens.calculate_merging_stats(None, use_weights='U')[0]
        x_stats = intens.calculate_merging_stats(None, use_weights='X')[0]
        self.assertEqual(y_stats.all_refl, 12)
        self.assertEqual(y_stats.unique_refl, 2)
        self.assertEqual(y_stats.stats_refl, 2)
        # value from the wiki and from 'xdscc12 -w'
        self.assertAlmostEqual(u_stats.cc_half(), 0.94582, delta=5e-6)
        # values from "mrfana -noice -n 1"
        self.assertAlmostEqual(y_stats.r_merge(), 0.3214, delta=5e-5)
        self.assertAlmostEqual(y_stats.r_meas(), 0.3520, delta=5e-5)
        self.assertAlmostEqual(y_stats.r_pim(), 0.1437, delta=5e-5)
        # values from iotbx.merging_statistics n_bins=1
        self.assertAlmostEqual(x_stats.r_merge(), 0.312, delta=5e-4)
        self.assertAlmostEqual(x_stats.r_meas(), 0.342, delta=5e-4)
        self.assertAlmostEqual(x_stats.r_pim(), 0.139, delta=5e-4)
        intens.merge_in_place(gemmi.DataType.Mean)
        self.assertEqual(len(intens), 2)

class TestConversion(unittest.TestCase):
    def test_4aap(self):
        def check_metadata(o, d):
            self.assertEqual(o.spacegroup.hm, 'P 32 2 1')
            self.assertEqual(o.cell.a, 68.575)
            self.assertEqual(d.cell.gamma, 120)
            self.assertEqual(d.wavelength, 0.97)
        doc = gemmi.cif.read(full_path('4aap-sf-subset.cif'))
        (rblock,) = gemmi.as_refln_blocks(doc)
        check_metadata(rblock, rblock)
        self.assertTrue(rblock.is_merged())
        mtz = gemmi.CifToMtz().convert_block_to_mtz(rblock)
        check_metadata(mtz, mtz.datasets[1])
        cif_string = gemmi.MtzToCif().write_cif_to_string(mtz)
        doc_out = gemmi.cif.read_string(cif_string)
        (rblock_out,) = gemmi.as_refln_blocks(doc_out)
        self.assertEqual(rblock.default_loop.tags[3:],
                         rblock_out.default_loop.tags)
        check_metadata(rblock_out, rblock_out)
        def cif_floats(rb, tag):
            return [float(x) if x != '?' else 8008.8008  # (arbitrary number)
                    for x in rb.block.find_values(tag)]
        tag = '_refln.phase_calc'
        self.assertEqual(cif_floats(rblock, tag), cif_floats(rblock_out, tag))
        tag = '_refln.pdbx_HL_A_iso'
        self.assertEqual(cif_floats(rblock, tag), cif_floats(rblock_out, tag))
        d1 = mtz.row_as_dict((7, 5, -1))
        self.assertEqual(d1['K'], 5)  # just to check row_as_dict()
        self.assertAlmostEqual(d1['FC'], 22.7, delta=1e-5)
        self.assertAlmostEqual(d1['PHIC'], 91.1, delta=1e-5)

        # test phase shift
        mtz.expand_to_p1()
        self.assertEqual(mtz.row_as_dict((7, 5, -1)), d1)
        mtz.ensure_asu()
        # (75-1) is not in P1 asu, we get Friedel mate
        self.assertEqual(mtz.row_as_dict((7, 5, -1)), {})
        d2 = mtz.row_as_dict((-7, -5, 1))
        self.assertAlmostEqual(d2['FC'], 22.7, delta=1e-5)
        self.assertAlmostEqual(d2['PHIC'], 360-91.1, delta=1e-5)
        self.assertAlmostEqual(d2['HLA'], d1['HLA'], delta=1e-5)
        self.assertAlmostEqual(d2['HLB'], -d1['HLB'], delta=1e-5)
        self.assertAlmostEqual(d2['HLC'], d1['HLC'], delta=1e-5)
        self.assertAlmostEqual(d2['HLD'], -d1['HLD'], delta=1e-5)
        # Expanding mtz file with SFTOOLS from CCP4 (command EXPAND)
        # and cctbx (miller_array.expand_to_p1().map_to_asu()) gives
        # different phases (and H-L coefficients) for some reflections,
        # including (-12, 5, 1).
        # Here we agree with cctbx (which gives PHIC -28.9).
        d3 = mtz.row_as_dict((-12, 5, 1))
        self.assertAlmostEqual(d3['FC'], 22.7, delta=1e-5)
        self.assertAlmostEqual(d3['PHIC'], 360-28.9, delta=1e-5)
        self.assertAlmostEqual(d3['HLA'], -5.28884, delta=1e-5)
        self.assertAlmostEqual(d3['HLB'], 10.0395, delta=1e-4)
        self.assertAlmostEqual(d3['HLC'], 1.53099, delta=1e-5)
        self.assertAlmostEqual(d3['HLD'], 4.64824, delta=1e-5)

    # flake8: noqa: E501
    def test_history_parsing(self):
        h1 = ['From AIMLESS, version 0.8.2, run on  1/ 3/2025 at 17:08:11',
              'REBATCH: run at 13:22:06 on  8/ 5/17',
              'From POINTLESS, version 1.11.1, run on  8/ 5/2017 at 13:22:06',
              'From DIALS 1.5.1-gf72becc3-release, run on 08/05/2017 at 12:22:05']
        (aimless, dials, g) = gemmi.get_software_from_mtz_history(h1)
        def check_sw(item, name, ver, classification):
            self.assertEqual(item.name, name)
            self.assertEqual(item.version, ver)
            self.assertEqual(item.classification, classification)
        classif = gemmi.SoftwareItem.Classification
        check_sw(aimless, 'AIMLESS', '0.8.2', classif.DataScaling)
        check_sw(dials, 'DIALS', '1.5.1-gf72becc3-release', classif.DataReduction)
        self.assertEqual(g.name, 'gemmi')
        self.assertEqual(g.classification, classif.DataExtraction)
        h2 = ['From XDS VERSION Jan 31, 2020  BUILT=20200417, run on 03/11/2020 at 17:27:55',
              'From POINTLESS, version 1.12.2, run on  3/11/2020 at 18:27:54']
        (xds, g) = gemmi.get_software_from_mtz_history(h2)
        check_sw(xds, 'XDS', 'Jan 31, 2020', classif.DataReduction)
        h3 = ['From STARANISO version: 2.3.80 (21-Oct-2021) on Sat, 23 Oct 2021 17:11:15 +0200.',
              "Signal_type = 'local <I/sigmaI>', signal_threshold = 1.20,",
              'averaging_radius = 0.1055/Ang.',
              'B=(  7.2464, 20.9790,  1.3204,  0.0000, -3.0933,  0.0000); g=6.06746E-01',
              'From XIA2 0.6.475-g7ac7bb6b-dials-2.2, run on 06/04/2021 at 19:34:04',
              'From FREERFLAG  6/ 4/2021 21:34:03 with fraction 0.050',
              'From FREERFLAG  6/ 4/2021 21:34:03 with fraction 0.050',
              'cctbx.french_wilson analysis',
              'From AIMLESS, version 0.7.4, run on  6/ 4/2021 at 21:33:58',
              'From SORTMTZ  6/ 4/2021 21:33:55  using keys: H K L M/ISYM BATCH',
              'From XDS VERSION Feb 5, 2021  BUILT=20210323, run on 06/04/2021 at 19:33:54',
              'From POINTLESS, version 1.12.8, run on  6/ 4/2021 at 21:33:53']
        (aimless, xds, g) = gemmi.get_software_from_mtz_history(h3)
        check_sw(xds, 'XDS', 'Feb 5, 2021', classif.DataReduction)
        check_sw(aimless, 'AIMLESS', '0.7.4', classif.DataScaling)
        h4 = ['From SCALA: run at 16:32:41 on 10/ 8/10']
        (scala, g) = gemmi.get_software_from_mtz_history(h4)
        check_sw(scala, 'SCALA', '', classif.DataScaling)
        h5 = ['From FREERFLAG, 21/ 3/97 12:34:50 with fraction 0.050',
              'From MTZUTILS, 21/ 3/97 12:34:45 after history:',
              'From TRUNCATE, 21/ 3/97 12:34:44',
              'SCALA: run at 12:34:33 on 21/ 3/97',
              'From SORTMTZ, 21/ 3/97 10:36:28  using keys: H K L M/ISYM BATCH',
              'From REINDEX, 21/ 3/97 10:36:25',
              'From SORTMTZ, 21/ 3/97 10:36:23  using keys: H K L M/ISYM BATCH',
              'From MOSFLM run on 20/ 3/97                                                    \x00',
              'From TRUNCATE, 21/ 3/97 12:34:44',
              'SCALA: run at 12:34:33 on 21/ 3/97',
              'From SORTMTZ, 21/ 3/97 10:36:28  using keys: H K L M/ISYM BATCH',
              'From REINDEX, 21/ 3/97 10:36:25',
              'From SORTMTZ, 21/ 3/97 10:36:23  using keys: H K L M/ISYM BATCH',
              'From MOSFLM run on 20/ 3/97                                                    \x00',
              'data from CAD on 21/ 3/97                                                      \x00']
        (scala, mosflm, g) = gemmi.get_software_from_mtz_history(h5)
        check_sw(scala, 'SCALA', '', classif.DataScaling)
        check_sw(mosflm, 'MOSFLM', '', classif.DataReduction)

class TestReciprocalGrid(unittest.TestCase):
    @unittest.skipIf(numpy is None, 'requires NumPy')
    def test_array_conversion(self):
        grid = gemmi.ReciprocalComplexGrid(4, 4, 4)
        expected_dtype = numpy.complex64
        self.assertEqual(grid.array.dtype, expected_dtype)
        new_val = 0+0j
        if numpy.__version__[:2] == '1.':
            possible_copy_values = [True, False]
        else:
            possible_copy_values = [None, True, False]
        for dtype in [None, numpy.complex64]:
            for copy in possible_copy_values:
                arr = numpy.array(grid, dtype=dtype, copy=copy)
                self.assertEqual(arr.dtype, expected_dtype,
                                 msg=f'for {dtype=} {copy=} {new_val=}, {grid.array[0,0,0]=}')
                new_val += 1+1j
                arr[0,0,0] = new_val
                grid_changed = (grid.array[0,0,0] == new_val)
                self.assertEqual(grid_changed, not copy,
                                 msg=f'for {dtype=} {copy=} {new_val=}, {grid.array[0,0,0]=}')

if __name__ == '__main__':
    unittest.main()
